/*
	CFLASH.CPP
	CompactFlash/FAT emulation routines for Dualis
	/Mic, 2006

	Logical memory layout:

	----------------------
	|        MBR         |
	----------------------
	|        FAT         |
	----------------------
	|    Root entries    |
	----------------------
	|   Subdirectories   |
	----------------------
	|     File data      |
	----------------------
	
*/

#include "cflash.h"


#define SECPERFAT	128
#define SECPERCLUS	16
#define MAXFILES	32768
#define SECRESV		2
#define NUMSECTORS	0x80000
#define NUMCLUSTERS	(NUMSECTORS/SECPERCLUS)



unsigned short				cf_reg_data,cf_reg_err,cf_reg_sec,cf_reg_lba1,cf_reg_lba2,
							cf_reg_lba3,cf_reg_lba4,cf_reg_cmd,cf_reg_sts;
unsigned int				CF_REG_DATA,CF_REG_ERR,CF_REG_SEC,CF_REG_LBA1,CF_REG_LBA2,
							CF_REG_LBA3,CF_REG_LBA4,CF_REG_CMD,CF_REG_STS;

BOOT_RECORD					MBR;
DIR_ENT						*files,*dirEntries,**dirEntryPtr;
FILE_INFO					*fileLink,*dirEntryLink;
UINT						currLBA;
UINT						filesysFAT,filesysRootDir,filesysData;
u16							FAT16[SECPERFAT*256];
int							numFiles,fileLevel,maxLevel,dirNum,numRootFiles;
int							*dirEntriesInCluster,clusterNum,
							firstDirEntCluster,lastDirEntCluster,
							lastFileDataCluster;
HANDLE						houtF;
char						*sRomPath;
int							activeDirEnt=-1;
UINT						bufferStart;
UINT						fileStartLBA,fileEndLBA;
unsigned short				freadBuffer[256];
DWORD						dwBytesRead;
HANDLE						hFile;
char						fpath[MAX_PATH+1];
bool						cflashDeviceEnabled = false;



int lfn_checksum() {
	int i;
	unsigned char chk;

	chk = 0;
	for (i=0; i < 11; i++) {
		chk = ((chk & 1) ? 0x80 : 0) + (chk >> 1) + (i < 8 ? files[numFiles].name[i] : files[numFiles].ext[i - 8]);
	}
	return chk;
}


const int lfnPos[13] = {1,3,5,7,9,14,16,18,20,22,24,28,30};


/* Add a DIR_ENT for the files */
void add_file(char *fname, WIN32_FIND_DATA *FindFileData) {
	int i,j,k,n;
	unsigned char chk;
	char *p;

	if (numFiles < MAXFILES-1) {
		if (strcmp(fname,"..") != 0) {
			for (i=0; i<strlen(fname); i++)
				if (fname[i]=='.') break;
			if ((i==0)&&(strcmp(fname,".")==0)) i = 1;
			for (j=0; j<i; j++)
				files[numFiles].name[j] = fname[j];
			for (; j<8; j++)
				files[numFiles].name[j] = 0x20;
			for (j=0; j<3; j++) {
				if ((j+i+1)>=strlen(fname)) break;
				files[numFiles].ext[j] = fname[j+i+1];
			}
			for (; j<3; j++)
				files[numFiles].ext[j] = 0x20;

			// See if LFN entries need to be added
			if (strlen(FindFileData->cAlternateFileName)>0) {
				chk = lfn_checksum();
				k = (strlen(FindFileData->cFileName)/13) + (((strlen(FindFileData->cFileName)%13)!=0)?1:0);
				numFiles += k;
			
				n = 0;
				j = 13;
				for (i=0; i<strlen(FindFileData->cFileName); i++) {
					if (j == 13) {
						n++;
						p = (char*)&files[numFiles-n].name[0];
						p[0] = n;
						p[0xB] = ATTRIB_LFN;
						p[0xD] = chk;
						j = 0;
					}
					*(p + lfnPos[j]) = FindFileData->cFileName[i];
					*(p + lfnPos[j]+1) = 0;
					j++;
				}
				for (; j<13; j++) {
					*(p + lfnPos[j]) = FindFileData->cFileName[i];
					*(p + lfnPos[j]+1) = 0;
				}
				p[0] |= 0x40;	// END
				for (i=0; i<strlen(fname); i++)
					if (fname[i]=='.') break;
				if ((i==0)&&(strcmp(fname,".")==0)) i = 1;
				for (j=0; j<i; j++)
					files[numFiles].name[j] = fname[j];
				for (; j<8; j++)
					files[numFiles].name[j] = 0x20;
				for (j=0; j<3; j++) {
					if ((j+i+1)>=strlen(fname)) break;
					files[numFiles].ext[j] = fname[j+i+1];
				}
				for (; j<3; j++)
					files[numFiles].ext[j] = 0x20;
			}

			files[numFiles].fileSize = FindFileData->nFileSizeLow;

			if (FindFileData->dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY) {
				if (strcmp(fname,".")==0)
					fileLink[numFiles].level = maxLevel;
				else
					fileLink[numFiles].level = maxLevel+1;
				files[numFiles].attrib = ATTRIB_DIR;
			} else {
				files[numFiles].attrib = 0;
			}

			fileLink[numFiles].parent = maxLevel * ((fileLevel>0)?1:0);	

			numFiles++;
		} else if (fileLevel > 0) {
			strncpy((char*)&files[numFiles].name[0],"..      ",8);
			strncpy((char*)&files[numFiles].ext[0],"   ",3);
			fileLink[numFiles].parent = maxLevel * ((fileLevel>0)?1:0);	
			files[numFiles].attrib = 0x10;
			numFiles++;
		}
	}
}


/* List all files and subdirectories recursively */
void list_files(char *path) {
	WIN32_FIND_DATA	FindFileData;
	HANDLE			hFind = INVALID_HANDLE_VALUE;
	char			DirSpec[MAX_PATH + 1],SubDir[MAX_PATH+1],fnameu[256];
	DWORD			dwError;
	char			*fname;
	int				i,j;

	fileLevel++;
	maxLevel++;
	
	strncpy(DirSpec, path, strlen(path)+1);
	strncat(DirSpec, "\\*", 3);

	hFind = FindFirstFile(DirSpec, &FindFileData);

	if (hFind == INVALID_HANDLE_VALUE) {
		wsprintf(buffer,"Invalid file handle. Error is %u\n", GetLastError());
		WriteConsole(hout,buffer,lstrlen(buffer),(unsigned long*)&dummy,NULL);
		return;
	} else {
		fname = (strlen(FindFileData.cAlternateFileName)>0)?FindFileData.cAlternateFileName:FindFileData.cFileName;
		strcpy(fnameu,fname);
		strupr(fnameu);
		add_file(fnameu,&FindFileData);

		while (FindNextFile(hFind, &FindFileData) != 0) {
			fname = (strlen(FindFileData.cAlternateFileName)>0)?FindFileData.cAlternateFileName:FindFileData.cFileName;
			strcpy(fnameu,fname);
			strupr(fnameu);
			
			add_file(fnameu,&FindFileData);

			if (numFiles==MAXFILES-1) break;

			if ((FindFileData.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY)&&
			   (fname[0] != '.')) {
				wsprintf(SubDir,"%s\\%s",path,fname);
				list_files(SubDir);
			}
		}
    
		dwError = GetLastError();
		FindClose(hFind);
		if (dwError != ERROR_NO_MORE_FILES) {
			wsprintf(buffer,"FindNextFile error. Error is %u\n", dwError);
			WriteConsole(hout,buffer,lstrlen(buffer),(unsigned long*)&dummy,NULL);
			return;
		}
	}
	if (numFiles < MAXFILES) {
		fileLink[numFiles].parent = maxLevel * ((fileLevel>0)?1:0);
		files[numFiles++].name[0] = 0;
	}
	fileLevel--;
}






/* Set up the MBR, FAT and DIR_ENTs */
bool cflash_build_fat() {
	int i,j,k,l,clust,baseCluster,numClusters,rootCluster;

	numFiles  = 0;
	fileLevel = -1;
	maxLevel  = -1;

	sRomPath  = (char*)zombie(ASK,ZOMBIE_ROM_PATH,0L,0L);
	files     = (DIR_ENT*)LocalAlloc(LPTR,MAXFILES*sizeof(DIR_ENT));
	fileLink  = (FILE_INFO*)LocalAlloc(LPTR,MAXFILES*sizeof(FILE_INFO));
	if ((files == NULL) || (fileLink == NULL))
		return false;

	list_files(sRomPath);

	wsprintf(buffer,"Found %d files\n",numFiles);
	WriteConsole(hout,buffer,lstrlen(buffer),(unsigned long*)&dummy,NULL);

	k            = 0;
	clusterNum   = rootCluster = (SECRESV + SECPERFAT)/SECPERCLUS;
	numClusters  = 0;
	clust        = 0;
	numRootFiles = 0;

	// Allocate memory to hold information about the files 
	dirEntries          = (DIR_ENT*)LocalAlloc(LPTR,numFiles*sizeof(DIR_ENT));
	dirEntryLink        = (FILE_INFO*)LocalAlloc(LPTR,numFiles*sizeof(FILE_INFO));
	dirEntriesInCluster = (int*)LocalAlloc(LPTR,NUMCLUSTERS*sizeof(int));
	dirEntryPtr         = (DIR_ENT**)LocalAlloc(LPTR,NUMCLUSTERS*sizeof(DIR_ENT*));
	if ((dirEntries==NULL) || (dirEntriesInCluster==NULL) || (dirEntryPtr==NULL))
		return false;

	for (i=0; i<NUMCLUSTERS; i++) {
		dirEntriesInCluster[i] = 0;
		dirEntryPtr[i] = NULL;
	}

	// Change the hierarcical layout to a flat one 
	for (i=0; i<=maxLevel; i++) {
		for (j=0; j<numFiles; j++) {
			if (fileLink[j].parent == i) {
				if (dirEntryPtr[clusterNum] == NULL)
					dirEntryPtr[clusterNum] = &dirEntries[k];
				dirEntryLink[k] = fileLink[j];
				dirEntries[k++] = files[j];
				if (files[j].attrib & ATTRIB_DIR) {
					if (strncmp((char*)&files[j].name[0],".       ",8)==0) {
						dirEntries[k-1].startCluster = dirEntryLink[k-1].level; //clusterNum;
					} else if (strncmp((char*)&files[j].name[0],"..      ",8)==0) {
						dirEntries[k-1].startCluster = dirEntryLink[k-1].parent; //clusterNum-1;
					} else {
						clust++;
						dirEntries[k-1].startCluster = clust;
					}
				} else {
					dirEntries[k-1].startCluster = clusterNum;
				}
				if (i==0) numRootFiles++;
				dirEntriesInCluster[clusterNum]++;
			}
		}
		clusterNum++;
		numClusters++;
	}

	// Free the file indexing buffer
	LocalFree((HLOCAL)files);
	LocalFree((HLOCAL)fileLink);

	// Set up the MBR
	MBR.bytesPerSector = 512;
	MBR.numFATs = 1;
	strcpy((char*)&MBR.OEMName[0],"DUALIS");
	strcpy((char*)&MBR.fat16.fileSysType[0],"FAT16  ");
	MBR.reservedSectors = SECRESV;
	MBR.numSectors = 524288;
	MBR.numSectorsSmall = 0;
	MBR.sectorsPerCluster = SECPERCLUS;
	MBR.sectorsPerFAT = SECPERFAT;
	MBR.rootEntries = 512;
	MBR.fat16.signature = 0xAA55;

	filesysFAT = 0 + MBR.reservedSectors;
	filesysRootDir = filesysFAT + (MBR.numFATs * MBR.sectorsPerFAT);
	filesysData = filesysRootDir + ((MBR.rootEntries * sizeof(DIR_ENT)) / 512);

	// Set up the cluster values for all subdirectories
	clust = filesysData / SECPERCLUS;
	firstDirEntCluster = clust;
	for (i=1; i<numFiles; i++) {
		if ((dirEntries[i].attrib & ATTRIB_DIR)!=0) {
			if (dirEntries[i].startCluster > rootCluster)
				dirEntries[i].startCluster += clust-rootCluster;
		}
	}
	lastDirEntCluster = clust+numClusters-1;

	// Set up the cluster values for all files
	clust += numClusters; //clusterNum;
	for (i=0; i<numFiles; i++) {
		if ((dirEntries[i].attrib & ATTRIB_DIR)==0) {
			dirEntries[i].startCluster = clust;
			clust += (dirEntries[i].fileSize/(512*SECPERCLUS));
			clust++;
		}
	}
	lastFileDataCluster = clust-1;

	// Set up the FAT16
	k = 2;
	for (i=0; i<numFiles; i++) {
		j = 0;
		l = dirEntries[i].fileSize - (512*SECPERCLUS);
		while (l > 0) {
			if (dirEntries[i].startCluster+j < MAXFILES)
				FAT16[dirEntries[i].startCluster+j] = dirEntries[i].startCluster+j+1;
			j++;
			l -= (512*16);
		} 
		if ((dirEntries[i].attrib & ATTRIB_DIR)==0) {
			if (dirEntries[i].startCluster+j < MAXFILES)
				FAT16[dirEntries[i].startCluster+j] = 0xFFFF;
		}
		k = dirEntries[i].startCluster+j;
	}

	/*for (i=0; i<numFiles; i++) {
		wsprintf(buffer,"[%s][%s] %d, cluster=%d\n\n",dirEntries[i].name,dirEntries[i].ext,dirEntries[i].cDate,dirEntries[i].startCluster);
		WriteConsole(hout,buffer,lstrlen(buffer),(unsigned long*)&dummy,NULL);
	}*/

	return true;
}




bool cflash_init() {
	cflashDeviceEnabled = false;
	currLBA = 0;
	if (activeDirEnt != -1)
		CloseHandle(hFile);
	activeDirEnt = -1;
	fileStartLBA = fileEndLBA = 0xFFFFFFFF;
	if (!cflash_build_fat())
		return false;
	cf_reg_sts = 0x58;	// READY

	cflashDeviceEnabled = true;
	return true;
}


/* Convert a space-padded 8.3 filename into an asciiz string */
void fatstring_to_asciiz(int dirent,char *out) {
	int i,j;

	for (i=0; i<8; i++) {
		if (dirEntries[dirent].name[i] == ' ') break;
		out[i] = dirEntries[dirent].name[i];
	}
	if ((dirEntries[dirent].attrib & 0x10)==0) {
		out[i++] = '.';
		for (j=0; j<3; j++) {
			if (dirEntries[dirent].ext[j] == ' ') break;
			out[i++] = dirEntries[dirent].ext[j];
		}
	}
	out[i] = '\0';
}



/* Resolve the path of a files by working backwards through the directory entries */
void resolve_path(int dirent) {
	int i;
	char dirname[128];

	while (dirEntryLink[dirent].parent > 0) {
		for (i=0; i<dirent; i++) {
			if ((dirEntryLink[dirent].parent==dirEntryLink[i].level) &&
				((dirEntries[i].attrib&ATTRIB_DIR)!=0)) {
				fatstring_to_asciiz(i,dirname);
				strcat(fpath,dirname);
				strcat(fpath,"\\");
				dirent = i;
				break;
			}
		}
	}
}


/* Read from a file using a 512 byte buffer */
unsigned short fread_buffered(int dirent,UINT cluster,UINT offset) {
	char fname[32];
	int i,j;

	offset += cluster*512*SECPERCLUS;

	if (dirent == activeDirEnt) {
		if ((offset < bufferStart) || (offset >= bufferStart + 512)) {
			SetFilePointer(hFile,offset,NULL,FILE_BEGIN);
			ReadFile(hFile,&freadBuffer,512,&dwBytesRead,NULL);
			bufferStart = offset;
		}
		return freadBuffer[(offset-bufferStart)>>1];
	}
	if (activeDirEnt != -1)
		CloseHandle(hFile);

	strcpy(fpath,sRomPath);
	strcat(fpath,"\\");
	
	resolve_path(dirent);

	fatstring_to_asciiz(dirent,fname);
	strcat(fpath,fname);

	wsprintf(buffer,"Opening [%s]\n",fpath);
	WriteConsole(hout,buffer,lstrlen(buffer),(unsigned long*)&dummy,NULL);

	hFile = CreateFile(fpath, GENERIC_READ, 0, NULL, OPEN_EXISTING,
	                   FILE_ATTRIBUTE_NORMAL|FILE_FLAG_SEQUENTIAL_SCAN,
		               NULL);
	if (hFile == INVALID_HANDLE_VALUE)
		return 0;
	ReadFile(hFile,&freadBuffer,512,&dwBytesRead,NULL);

	bufferStart = offset;
	activeDirEnt = dirent;
	fileStartLBA = (dirEntries[i].startCluster*512*SECPERCLUS);
	fileEndLBA = fileStartLBA + dirEntries[i].fileSize;

	return freadBuffer[(offset-bufferStart)>>1];
}



/* Read from one of the CF device registers */
unsigned int cflash_read(unsigned int address) {
	unsigned char *p;
	unsigned short s;
	int i;
	UINT cluster,cluster2,fileLBA;

	if (address == CF_REG_STS) {
		s = cf_reg_sts;
		return s; 

	} else if (address == CF_REG_DATA) {
		cluster = (currLBA / (512 * SECPERCLUS));
		cluster2 = (((currLBA/512) - filesysData) / SECPERCLUS) + 2;

		if (cf_reg_cmd == 0x20) {
			// Reading from the MBR 
			if (currLBA < 512) {
				p = (unsigned char*)&MBR;
				s = *(unsigned short *)(p + currLBA);

			// Reading the FAT 
			} else if ((currLBA >= filesysFAT*512) && (currLBA < filesysRootDir*512)) {
				p = (unsigned char*)&FAT16[0];
				s = *(unsigned short *)(p + (currLBA-filesysFAT*512));

			// Reading directory entries 
			} else if ((currLBA >= filesysRootDir*512) &&
				       (cluster <= lastDirEntCluster) && //(cluster < firstDirEntCluster) &&
					   ((currLBA-(((cluster-(filesysRootDir/SECPERCLUS))*SECPERCLUS+filesysRootDir)*512)) < (dirEntriesInCluster[cluster]*32))) {
				p = (unsigned char*)dirEntryPtr[cluster];
				s = *(unsigned short*)(p + (currLBA-(((cluster-(filesysRootDir/SECPERCLUS))*SECPERCLUS+filesysRootDir)*512)));

			// Reading file data 
			} else if ((cluster2 > lastDirEntCluster) && (cluster2 <= lastFileDataCluster)) { //else if ((cluster>lastDirEntCluster)&&(cluster<=lastFileDataCluster)) {
				fileLBA = currLBA - (filesysData-32)*512;	// 32 = # sectors used for the root entries

				//wsprintf(buffer,"%d, %d [%d, %d, %d]\n",fileLBA,currLBA,firstDirEntCluster,lastDirEntCluster,lastFileDataCluster);
				//WriteConsole(hout,buffer,lstrlen(buffer),(unsigned long*)&dummy,NULL);

				// Check if the read is from the currently opened file 
				if ((fileLBA >= fileStartLBA) && (fileLBA < fileEndLBA)) {
					cluster = (fileLBA / (512 * SECPERCLUS));
					s = fread_buffered(activeDirEnt,cluster-dirEntries[activeDirEnt].startCluster,fileLBA&((512*SECPERCLUS)-1)); 
				} else {
					for (i=0; i<numFiles; i++) {
						if ((fileLBA>=(dirEntries[i].startCluster*512*SECPERCLUS)) &&
							(fileLBA <(dirEntries[i].startCluster*512*SECPERCLUS)+dirEntries[i].fileSize)) {
							cluster = (fileLBA / (512 * SECPERCLUS));
							s = fread_buffered(i,cluster-dirEntries[i].startCluster,fileLBA&((512*SECPERCLUS)-1));
							break;
						}
					}
				}
			} 
			currLBA += 2;
			return s;
		}

	} else if (address == CF_REG_CMD) {
		s = 0;
		return 0;

	} else if (address == CF_REG_LBA1) {
		s = cf_reg_lba1;
		return s; 
	}
}


/* Write to one of the CF device registers */
void cflash_write(unsigned int address,unsigned int data) {
	if (address==CF_REG_STS) {
		cf_reg_sts = data&0xFFFF;

	} else if (address==CF_REG_CMD) {
		cf_reg_cmd = data&0xFF;
		cf_reg_sts = 0x58;	// READY
	
	} else if (address==CF_REG_LBA1) {
		cf_reg_lba1 = data&0xFF;
		currLBA = (currLBA&0xFFFFFF00)|cf_reg_lba1;
	
	} else if (address==CF_REG_LBA2) {
		cf_reg_lba2 = data&0xFF;
		currLBA = (currLBA&0xFFFF00FF)|(cf_reg_lba2<<8);
	
	} else if (address==CF_REG_LBA3) {
		cf_reg_lba3 = data&0xFF;
		currLBA = (currLBA&0xFF00FFFF)|(cf_reg_lba3<<16);
	
	} else if (address==CF_REG_LBA4) {
		cf_reg_lba4 = data&0xFF;
		currLBA = (currLBA&0x00FFFFFF)|((cf_reg_lba4&0x0F)<<24);
		currLBA *= 512;
	}
}



void cflash_close() {
	if (cflashDeviceEnabled) {
		cflashDeviceEnabled = false;

		if (dirEntries != NULL)
			LocalFree((HLOCAL)dirEntries);

		if (dirEntryLink != NULL)
			LocalFree((HLOCAL)dirEntryLink);

		if (activeDirEnt != -1)
			CloseHandle(hFile);
	}
}
