// DS MMU emulator
// Mic, 2005/2006

#include <windows.h>
#include <commctrl.h>
#include <string.h>
#include "zombie.h"
#include "resource.h"
#include "arm7_read.h"
#include "arm7_write.h"
#include "arm9_read.h"
#include "arm9_write.h"
#include "cflash.h"
#include <stdio.h>



extern "C" void __stdcall _lldiv(void);
extern "C" void __stdcall _llrem(void);


touchpad					*touch;

// Timer prescalar
const unsigned char			prescalar[4] = {0,6,8,10};

// DMA maximum transfer counts
const int					dmamax[4] = {0x200000,0x200000,0x200000,0x200000};

const int					keymask[] = {0x040,0x080,0x020,0x010,0x001,0x002,0x200,0x100,0x008,0x004};
int							keycodes[] = {38,40,37,39,88,90,76,82,13,32};
int							btnXY[2];

HINSTANCE					hInst;
HANDLE						hout;

unsigned char				*ROM,*ROMsub,*SYSROM,*IWRAM,*EXRAM,*cart,*VRAM,*IOREGS,*mmu_memory,*BIOS,
							*ITCM,*DTCM,*OBJ_VRAM,*OBJ_S_VRAM,*LCDC_VRAM,*PLT_VRAM,*PLT_S_VRAM,
							*S_VRAM,*OAM,*EIWRAM,*SYSROM_ARM7,*IOREGS_ARM7,*CARTRAM;
unsigned char				*iomap,**mmap,*arm9_pageTable[256],*arm7_pageTable[256];
unsigned int				arm9_pageLength[16],arm9_pageMask[256],
							arm7_pageLength[16],arm7_pageMask[256];
int							*arm9_pageTableAddr  = (int*)&arm9_pageTable[0],
							*arm9_pageLengthAddr = (int*)&arm9_pageLength[0],
							*arm9_pageMaskAddr   = (int*)&arm9_pageMask[0],
							*arm7_pageTableAddr  = (int*)&arm7_pageTable[0],
							*arm7_pageLengthAddr = (int*)&arm7_pageLength[0],
							*arm7_pageMaskAddr   = (int*)&arm7_pageMask[0];
unsigned int				dtcmPage=0, dtcmBase = 0x800000, itcmBase=0;
dtcmBackup					dtcmLast;

int							arm9RecvFifo[16],arm9SendFifo[16],arm7RecvFifo[16],arm7SendFifo[16];
int							arm9RecvFifoPos,arm9SendFifoPos,arm7RecvFifoPos,arm7SendFifoPos;
unsigned short int			arm9FifoCnt,arm7FifoCnt;
char						ipcSync,ipcSync7;

bool						useExtBios = false;
char						szBiosName[] = "bios.bin";
HANDLE						hBios;

unsigned short int			keys,irqFlags,irqFlags7,irqAck,irqMask;
int							*arm9_regs,*arm7_regs;
int							romOffset,romSize;
int							iwramOffset,iwramSize;
int							exramOffset,exramSize;
int							vramOffset,vramSize;
char						mmuName[] = "DS MMU";
int							dummy;
unsigned int				*arm9_cycle,*arm7_cycle;
unsigned int				addressLatch;
bool						logTimer=false,logDMA=false;

char						*szRomPath;

mmu_read_byte_callback		arm9_rd8[256];
mmu_read_word_callback		arm9_rd16[256];
mmu_read_dword_callback		arm9_rd32[256];
mmu_write_byte_callback		arm9_wr8[256];
mmu_write_word_callback		arm9_wr16[256];
mmu_write_dword_callback	arm9_wr32[256];

mmu_read_byte_callback		arm7_rd8[256];
mmu_read_word_callback		arm7_rd16[256];
mmu_read_dword_callback		arm7_rd32[256];
mmu_write_byte_callback		arm7_wr8[256];
mmu_write_word_callback		arm7_wr16[256];
mmu_write_dword_callback	arm7_wr32[256];

mmu_read_word_callback		rdh_ipc,rdh_ipc_mirror;

cpu_irq_callback			cpu_irq;
cpu_irq_callback			arm7_irq;
cpu_dma_callback			arm9_dma;
cpu_dma_callback			arm7_dma;
gpu_set_callback			gpu_set;
gpu_get_callback			gpu_get;
apu_write_callback			apu_write;
zombie_callback				zombie;

char						buffer[256];
FILE						*log;
int							blockDMA=0;
dma_channel					dma[2][4];
timer_channel				timer[2][4];
SYSTEMTIME					*systime;

vram_bank					vramBank[9];
unsigned char				*vramBase[9];
unsigned char				*vramChunk[1024];
extern int					vramSettings[9][32][3];
unsigned char				*VRAM_A,*VRAM_B,*VRAM_C,*VRAM_D,*VRAM_E,*VRAM_F,*VRAM_G,*VRAM_H,*VRAM_I;

int							ipc_page = 0x20; //0x23;
int							ipc_x=4,ipc_y=6,ipc_z=20;
int							cfDevice=1;



// For ARM7/ARM9
unsigned int				sysromcode_0[8] = {
		0,			// 00000000
		0,			// 00000004
		0,			// 00000008
		0,			// 0000000C
		0,			// 00000010
		0,			// 00000014
		0xEA000042,	// 00000018: B $128
		0			// 0000001C
};

// For ARM9
unsigned int				sysromcode_128[9] = {
		0xE92D500F,	// 00000128: STMFD R13!,{R0-R3,R12,R14}

		0xee193f11,	// 12C: mrc 15,0,r3,c9,c1,{0}
		0xe1a03623, // 130: mov	r3, r3, lsr #12
		0xe1a03603, // 134: mov	r3, r3, lsl #12
		0xe2833901, // 138: add	r3, r3, #0x4000

		0xE28FE000,	// 0000013C: ADD R14,R15,#0
		0xE513F004,	// 00000140: LDR R15,[R3,-4]
		0xE8BD500F,	// 00000144: LDMFD R13!,{R0-R3,R12,R14}
		0xE25EF004	// 00000148: SUBS R15,R14,#4
};


// For ARM7
unsigned int				arm7_sysromcode_128[6] = {
		0xe92d500f,	// 128: stmdb   sp!, {r0, r1, r2, r3, ip, lr}
		0xe3a00301, // 12C: mov     r0, #0x4000000
		0xe28fe000, // 130: add     lr, pc, #0    
		0xe510f004, // 134: ldr     pc, [r0, #-4]
		0xe8bd500f, // 138: ldmia   sp!, {r0, r1, r2, r3, ip, lr}
		0xe25ef004  // 13C: subs    pc, lr, #4      ; 0x4
};



BOOL CALLBACK DlgProc(HWND hdlg, UINT msg, WPARAM wParam, LPARAM lParam);



bool __stdcall DllMain(HINSTANCE hDLL,DWORD reason,LPVOID unused) {
	if (reason==DLL_PROCESS_ATTACH)
		hInst = hDLL;

	return true;
}





///////////////////////////////////////////////////////////////////////////////////////


float fourgigs = 4294967296.0f;
int divNumer,divDenom;
int divNumer2,divDenom2;
int divNumer3,divDenom3;
int divNumer4,divDenom4;

/////////////////////////////////////////////////////////////////////////////////////////

// Co-processor division
void cp_div(short int cnt)
{
	static short int divCnt=CP_DIV0;
	__int64 bigNumer,bigDenom,bigQ,bigR;

	if (cnt != -1) divCnt = (divCnt&0xFFFC)|cnt;
	divCnt &= ~CP_BUSY;

	divNumer  = *(int*)(IOREGS + REG_NUMER_L);
	divNumer2 = *(int*)(IOREGS + REG_NUMER_L + 4);
	divDenom  = *(int*)(IOREGS + REG_DENOM_L);
	divDenom2  = *(int*)(IOREGS + REG_DENOM_L + 4);
	
	if (divDenom) {
		divCnt &= ~CP_DIV0;

		if ((divCnt&3)==0) {		// 32bit / 32bit => 32bit quotient, 32bit remainder
			__asm {
				mov eax,divNumer
				cdq
				idiv dword ptr [divDenom]
				mov divNumer,eax
				mov divDenom,edx
			}
			*(int*)(IOREGS + REG_DIVRESULT_L) = divNumer;
			*(int*)(IOREGS + REG_DIVREM_L) = divDenom;

		} else if ((divCnt&3)<3) {
			if ((divCnt&3)==1) {	// 64bit / 32bit => 64bit quotient, 32bit remainder
				divNumer3 = divNumer;
				divNumer4 = divNumer2;
			
				__asm {
					// Input:    64 bit dividend in EBX::ECX, 32 bit divisor in ESI
					// Output:   64 bit result in EBX::EAX, 32 bit remainder in EDX
					// Uses:     EAX, EBX, ECX, EDX, ESI
					mov divNumer3,0
					mov divNumer4,0

					mov ecx,[divNumer]
					mov ebx,[divNumer2]
					mov esi,[divDenom]
					test ebx,0x80000000	// Check sign of the dividend
					jz __numer_uns
						not ecx
						not ebx
						add ecx,1
						adc ebx,0
						mov divNumer3,1
					__numer_uns:
					test esi,0x80000000	// Check sign of the divisor
					jz __denom_uns
						neg esi
						mov divNumer4,1
					__denom_uns:

					// Divide high dword by divisor.
					mov     eax, ebx
					xor     edx, edx
					div     esi
					// Put remainder as high dword of the original dividend.
					mov     ebx, eax
					mov     eax, ecx
					div     esi

					test divNumer3,1
					jz __rem_pos
						neg edx
					__rem_pos:
					mov ecx,divNumer3
					xor ecx,divNumer4
					jz __quota_pos
						not eax
						not ebx
						add eax,1
						adc ebx,0
					__quota_pos:
					mov divNumer,eax
					mov divNumer2,ebx
					mov divDenom,edx
				}

				*(int*)(IOREGS + REG_DIVRESULT_L) = divNumer;
				*(int*)(IOREGS + REG_DIVRESULT_L+4) = divNumer2;
				*(int*)(IOREGS + REG_DIVREM_L) = divDenom;
				*(int*)(IOREGS + REG_DIVREM_L+4) = 0;

			} else if ((divCnt&3)==2) {
				__asm {
					// Not supported yet, cause an exception
					xor edi,edi
					div edi
				}
			}
		}
	} else
		divCnt |= CP_DIV0;

	*(short int*)(IOREGS + REG_DIVCNT) = divCnt;
}



// Co-processor square root
void cp_sqrt(short int cnt) {
	static short int sqrtCnt=0;
	int sqrtParm;

	if (cnt != -1) sqrtCnt |= cnt;
	sqrtCnt &= ~CP_BUSY;
	sqrtParm = *(int*)(IOREGS + REG_SQRTPARM);
	__asm {
		fild dword ptr [sqrtParm]
		fsqrt
		fwait
		fistp dword ptr [sqrtParm]
	}
	*(int*)(IOREGS + REG_SQRTRESULT) = sqrtParm;
	*(short int*)(IOREGS + REG_SQRTCNT) = sqrtCnt;
}



void ipc_fifo_recv(int proc) {
	wsprintf(buffer,"Processor %d receiving at %08X, fifocnt = %04X,%04X\n",
		proc,arm9_regs[15],arm9FifoCnt,arm7FifoCnt);
	WriteConsole(hout,buffer,lstrlen(buffer),(unsigned long*)&dummy,NULL);
	if (proc==0) {
		if (arm9RecvFifoPos<1)
			arm9FifoCnt |= 0x4100;
	} else {
		if (arm7RecvFifoPos<1)
			arm7FifoCnt |= 0x4100;
	}
	*(unsigned short int*)(arm9_pageTable[0x40] + 0x184) = arm9FifoCnt;
	*(unsigned short int*)(arm7_pageTable[0x40] + 0x184) = arm7FifoCnt;
}



void ipc_fifo_send(int proc,int dat) {
	if (proc==0) {
		if (arm9SendFifoPos<16) {
			arm9SendFifo[arm9SendFifoPos++] = dat;
			if (arm9SendFifoPos==16)
				arm9FifoCnt |= 2;
			arm9FifoCnt &= 0xFFFE;
			if (arm7RecvFifoPos<16) {
				arm7RecvFifo[arm7RecvFifoPos++] = dat;
				if (arm7RecvFifoPos==16)
					arm7FifoCnt |= 200;
				arm7FifoCnt &= 0xFEFF;
			} else {
				arm7FifoCnt |= 0x4200;
			}
		} else {
			arm9FifoCnt |= 0x4002;
		}
	} else {
		if (arm7SendFifoPos<16) {
			arm7SendFifo[arm7SendFifoPos++] = dat;
			if (arm7SendFifoPos==16)
				arm7FifoCnt |= 2;
			arm7FifoCnt &= 0xFFFE;
			if (arm9RecvFifoPos<16) {
				arm9RecvFifo[arm9RecvFifoPos++] = dat;
				if (arm9RecvFifoPos==16)
					arm9FifoCnt |= 200;
				arm9FifoCnt &= 0xFEFF;
			} else {
				arm9FifoCnt |= 0x4200;
			}
		} else {
			arm7FifoCnt |= 0x4002;
		}
	}
	wsprintf(buffer,"Processor %d sent %08X at %08X, fifocnt = %04X,%04X\n",
		proc,dat,arm9_regs[15],arm9FifoCnt,arm7FifoCnt);
	WriteConsole(hout,buffer,lstrlen(buffer),(unsigned long*)&dummy,NULL);

	*(unsigned short int*)(arm9_pageTable[0x40] + 0x184) = arm9FifoCnt;
	*(unsigned short int*)(arm7_pageTable[0x40] + 0x184) = arm7FifoCnt;
}


void remap_vram(int bank,int setting)
{
	int i,j,k;

	if (vramBank[bank].type != DISABLED) {
		j = vramBank[bank].base;
		if (j!=-1) {
			for (i=0; i<vramBank[bank].size; i++)
				vramChunk[j+i] = 0;
		}
		for (k=0; k<9; k++)
			if ((k!=bank)&&(vramBank[k].base == j)) break;
		if ((k<9)&&(j!=-1)) {
			wsprintf(buffer,"Bank %c has the same base as bank %c (%X)\n",
				'A'+(char)k,'A'+(char)bank,vramBank[k].base);
			WriteConsole(hout,buffer,lstrlen(buffer),(unsigned long*)&dummy,NULL);
			for (i=0; i<vramBank[k].size; i++)
				vramChunk[j+i] = vramBase[k] + (i<<14);
		}
	}

	if ((setting&0x80)==0) {
		vramBank[bank].type = DISABLED;
		return;
	}

	setting ^= 0x80;	// Clear bit 7

	vramBank[bank].size = vramSettings[bank][setting][1];
	vramBank[bank].type = vramSettings[bank][setting][2];
	if (vramSettings[bank][setting][0] != -1) {
		vramBank[bank].base = vramSettings[bank][setting][0]>>2;
	} else {
		vramBank[bank].base = -1;
		return;
	}

	if (vramBank[bank].type != DISABLED) {
		/*wsprintf(buffer,"Bank %c: base=%X, size=%X, type=%d\n",
			'A'+(char)bank,vramBank[bank].base,vramBank[bank].size,vramBank[bank].type);
		WriteConsole(hout,buffer,lstrlen(buffer),(unsigned long*)&dummy,NULL);*/

		j = vramBank[bank].base;
		for (k=0; k<bank; k++)
			if (vramBank[k].base == j) break;
		if (k==bank) {
			for (i=0; i<vramBank[bank].size; i++)
				vramChunk[j+i] = vramBase[bank] + (i<<14);
		}
	} 
}



void remap_wram(int proc) {
	int i;

	wsprintf(buffer,"Remapping WRAM banks with setting %d\n",
			IOREGS[0x247]);
	WriteConsole(hout,buffer,lstrlen(buffer),(unsigned long*)&dummy,NULL);

	switch (IOREGS[0x247]&3) { 
		case 0:
			for (i=0; i<8; i++) {
				arm9_pageTable[0x30+i] = arm9_pageTable[0x38+i] = IWRAM;
				arm9_rd8[0x30+i]  = arm9_rd8[0x38+i]  = (mmu_read_byte_callback)arm9_read_byte_wram;
				arm9_rd16[0x30+i] = arm9_rd16[0x38+i] = (mmu_read_word_callback)arm9_read_word_wram;
				arm9_rd32[0x30+i] = arm9_rd32[0x38+i] = (mmu_read_dword_callback)arm9_read_dword_wram;
				arm9_wr8[0x30+i]  = arm9_wr8[0x38+i]  = (mmu_write_byte_callback)arm9_write_byte_wram;
				arm9_wr16[0x30+i] = arm9_wr16[0x38+i] = (mmu_write_word_callback)arm9_write_word_wram;
				arm9_wr32[0x30+i] = arm9_wr32[0x38+i] = (mmu_write_dword_callback)arm9_write_dword_wram;
				arm9_pageMask[0x30+i] = arm9_pageMask[0x38+i] = 0x7FFF;

				arm7_pageTable[0x30+i] = EIWRAM;
				arm7_rd8[0x30+i]  = (mmu_read_byte_callback)arm7_read_byte_ewram;
				arm7_rd16[0x30+i] = (mmu_read_word_callback)arm7_read_word_ewram;
				arm7_rd32[0x30+i] = (mmu_read_dword_callback)arm7_read_dword_ewram;
				arm7_wr8[0x30+i]  = (mmu_write_byte_callback)arm7_write_byte_ewram;
				arm7_wr16[0x30+i] = (mmu_write_word_callback)arm7_write_word_ewram;
				arm7_wr32[0x30+i] = (mmu_write_dword_callback)arm7_write_dword_ewram;
				arm7_pageMask[0x30+i] = 0xFFFF;
			}
			break;
		case 1:
			for (i=0; i<8; i++) {
				arm9_pageTable[0x30+i] = arm9_pageTable[0x38+i] = IWRAM+0x4000;
				arm9_rd8[0x30+i]  = arm9_rd8[0x38+i]  = (mmu_read_byte_callback)arm9_read_byte_wram;
				arm9_rd16[0x30+i] = arm9_rd16[0x38+i] = (mmu_read_word_callback)arm9_read_word_wram;
				arm9_rd32[0x30+i] = arm9_rd32[0x38+i] = (mmu_read_dword_callback)arm9_read_dword_wram;
				arm9_wr8[0x30+i]  = arm9_wr8[0x38+i]  = (mmu_write_byte_callback)arm9_write_byte_wram;
				arm9_wr16[0x30+i] = arm9_wr16[0x38+i] = (mmu_write_word_callback)arm9_write_word_wram;
				arm9_wr32[0x30+i] = arm9_wr32[0x38+i] = (mmu_write_dword_callback)arm9_write_dword_wram;
				arm9_pageMask[0x30+i] = arm9_pageMask[0x38+i] = 0x3FFF;

				arm7_pageTable[0x30+i] = IWRAM;
				arm7_rd8[0x30+i]  = (mmu_read_byte_callback)arm7_read_byte_wram;
				arm7_rd16[0x30+i] = (mmu_read_word_callback)arm7_read_word_wram;
				arm7_rd32[0x30+i] = (mmu_read_dword_callback)arm7_read_dword_wram;
				arm7_wr8[0x30+i]  = (mmu_write_byte_callback)arm7_write_byte_wram;
				arm7_wr16[0x30+i] = (mmu_write_word_callback)arm7_write_word_wram;
				arm7_wr32[0x30+i] = (mmu_write_dword_callback)arm7_write_dword_wram;
				arm7_pageMask[0x30+i] = 0x3FFF;
			}
			break;
		case 2:
			for (i=0; i<8; i++) {
				arm9_pageTable[0x30+i] = arm9_pageTable[0x38+i] = IWRAM;
				arm9_rd8[0x30+i]  = arm9_rd8[0x38+i]  = (mmu_read_byte_callback)arm9_read_byte_wram;
				arm9_rd16[0x30+i] = arm9_rd16[0x38+i] = (mmu_read_word_callback)arm9_read_word_wram;
				arm9_rd32[0x30+i] = arm9_rd32[0x38+i] = (mmu_read_dword_callback)arm9_read_dword_wram;
				arm9_wr8[0x30+i]  = arm9_wr8[0x38+i]  = (mmu_write_byte_callback)arm9_write_byte_wram;
				arm9_wr16[0x30+i] = arm9_wr16[0x38+i] = (mmu_write_word_callback)arm9_write_word_wram;
				arm9_wr32[0x30+i] = arm9_wr32[0x38+i] = (mmu_write_dword_callback)arm9_write_dword_wram;
				arm9_pageMask[0x30+i] = arm9_pageMask[0x38+i] = 0x3FFF;

				arm7_pageTable[0x30+i] = IWRAM+0x4000;
				arm7_rd8[0x30+i]  = (mmu_read_byte_callback)arm7_read_byte_wram;
				arm7_rd16[0x30+i] = (mmu_read_word_callback)arm7_read_word_wram;
				arm7_rd32[0x30+i] = (mmu_read_dword_callback)arm7_read_dword_wram;
				arm7_wr8[0x30+i]  = (mmu_write_byte_callback)arm7_write_byte_wram;
				arm7_wr16[0x30+i] = (mmu_write_word_callback)arm7_write_word_wram;
				arm7_wr32[0x30+i] = (mmu_write_dword_callback)arm7_write_dword_wram;
				arm7_pageMask[0x30+i] = 0x3FFF;
			}
			break;
		default:
			for (i=0; i<8; i++) {
				arm9_pageTable[0x30+i] = IWRAM;
				arm9_rd8[0x30+i]  = arm9_rd8[0x38+i]  = (mmu_read_byte_callback)arm9_read_byte_zero;
				arm9_rd16[0x30+i] = arm9_rd16[0x38+i] = (mmu_read_word_callback)arm9_read_word_zero;
				arm9_rd32[0x30+i] = arm9_rd32[0x38+i] = (mmu_read_dword_callback)arm9_read_dword_zero;
				arm9_wr8[0x30+i]  = arm9_wr8[0x38+i]  = (mmu_write_byte_callback)arm9_write_byte_na;
				arm9_wr16[0x30+i] = arm9_wr16[0x38+i] = (mmu_write_word_callback)arm9_write_word_na;
				arm9_wr32[0x30+i] = arm9_wr32[0x38+i] = (mmu_write_dword_callback)arm9_write_dword_na;
				arm9_pageMask[0x30+i] = 0x0000;

				arm7_pageTable[0x30+i] = IWRAM;
				arm7_rd8[0x30+i]  = (mmu_read_byte_callback)arm7_read_byte_wram;
				arm7_rd16[0x30+i] = (mmu_read_word_callback)arm7_read_word_wram;
				arm7_rd32[0x30+i] = (mmu_read_dword_callback)arm7_read_dword_wram;
				arm7_wr8[0x30+i]  = (mmu_write_byte_callback)arm7_write_byte_wram;
				arm7_wr16[0x30+i] = (mmu_write_word_callback)arm7_write_word_wram;
				arm7_wr32[0x30+i] = (mmu_write_dword_callback)arm7_write_dword_wram;
				arm7_pageMask[0x30+i] = 0x7FFF;
			}
			break;
	}
	zombie(TELL,MMU_PAGE_TABLE,(void*)&arm9_pageTable[0],(void*)0L);
	zombie(TELL,MMU_PAGE_TABLE,(void*)&arm7_pageTable[0],(void*)1L);
}

/////////////////////////////////////////////////////////////////////////////////////////




bool __stdcall mmu_init(zombie_callback commlink) {
	int i;

	zombie = commlink;

	zombie(TELL,ZOMBIE_DEBUG_MSG,"MMU init",0);

	romOffset	= 0x8000000;
	romSize		= 0x8000000;

	exramOffset = 0x2000000;
	exramSize	= 0x400000;

	iwramOffset = 0x3000000;
	iwramSize	= 0x8000;

	vramOffset	= 0x6000000;
	vramSize	= 0x40000;

	mmu_memory = (unsigned char*)LocalAlloc(LPTR,iwramSize + exramSize + 0x100000 + 0x4000 +
		                                         0x4000 + 0x4000 + 0x8000 + 0x10000 + 0x4000 +
												 0x10000);
	IWRAM	= mmu_memory;
	EXRAM	= IWRAM + iwramSize;
	SYSROM	= EXRAM + exramSize;
	BIOS	= SYSROM + 0x100000; 
	EIWRAM	= BIOS+0x4000;
	DTCM	= EIWRAM + 0x10000;
	SYSROM_ARM7 = DTCM + 0x8000;
	CARTRAM = SYSROM_ARM7 + 0x4000;
	ITCM	= CARTRAM + 0x10000;
	

	if (mmu_memory == NULL)
		return false;

	if (useExtBios)	{
		if ((hBios = CreateFile(szBiosName,GENERIC_READ,0,NULL,OPEN_EXISTING,
						  FILE_ATTRIBUTE_NORMAL+FILE_FLAG_SEQUENTIAL_SCAN,NULL)) != INVALID_HANDLE_VALUE)
		{
			ReadFile(hBios,BIOS,0x4000,(LPDWORD)&i,NULL);
			if (i==0x4000)
				useExtBios = false;
		} else
			useExtBios = false;
	}


	for (i=0; i<0x10; i++) {
		arm9_pageTable[0x30+i] = IWRAM;
		arm9_pageTable[0x20+i] = EXRAM + ((i&3)*0x100000);
		arm9_pageTable[0x00+i] = DTCM; 
		arm9_pageTable[0xF0+i] = SYSROM;
		arm9_pageTable[0xA0+i] = CARTRAM;

		arm7_pageTable[0x00+i]     = SYSROM_ARM7;
		arm7_pageTable[0x30+(i&7)] = IWRAM;
		arm7_pageTable[0x38+(i&7)] = EIWRAM;
		arm7_pageTable[0x20+i]     = EXRAM + ((i&3)*0x100000);
		arm7_pageTable[0xA0+i]     = CARTRAM;
	}
	
	arm9_pageLength[0xF] = 0x100000;
	arm9_pageLength[3]	= 0x8000;
	arm9_pageLength[2]	= 0x100000; //0x400000;
	arm9_pageLength[0]	= 0x4000;
	arm9_pageLength[0xA] = 0x10000;

	arm7_pageLength[3]	= 0x8000;
	arm7_pageLength[2]	= 0x100000; //0x400000;
	arm7_pageLength[0]	= 0x4000;
	arm7_pageLength[0xA] = 0x10000;


	for (i=0; i<4; i++)	{
		dma[0][i].sad	 = 0xB0 + (i*0xC);
		dma[0][i].dad	 = 0xB4 + (i*0xC);
		dma[0][i].cnt_l = 0xB8 + (i*0xC);
		dma[0][i].cnt_h = 0xBA + (i*0xC);

		dma[1][i].sad	 = 0xB0 + (i*0xC);
		dma[1][i].dad	 = 0xB4 + (i*0xC);
		dma[1][i].cnt_l = 0xB8 + (i*0xC);
		dma[1][i].cnt_h = 0xBA + (i*0xC);

		timer[0][i].enabled = timer[0][i].unchanged = false;
		timer[0][i].tcurr   = 0;
		timer[0][i].tstart  = 0;
		timer[0][i].cntval  = 0;

		timer[1][i].enabled = timer[1][i].unchanged = false;
		timer[1][i].tcurr   = 0;
		timer[1][i].tstart  = 0;
		timer[1][i].cntval  = 0;
	}

	for (i=0; i<256; i++) {
		arm9_rd8[i]  = (mmu_read_byte_callback)arm9_read_byte_gen;
		arm9_rd16[i] = (mmu_read_word_callback)arm9_read_word_gen;
		arm9_rd32[i] = (mmu_read_dword_callback)arm9_read_dword_gen;
		arm9_wr8[i]  = (mmu_write_byte_callback)arm9_write_byte_gen;
		arm9_wr16[i] = (mmu_write_word_callback)arm9_write_word_gen;
		arm9_wr32[i] = (mmu_write_dword_callback)arm9_write_dword_gen;

		arm7_rd8[i]  = (mmu_read_byte_callback)arm7_read_byte_gen;
		arm7_rd16[i] = (mmu_read_word_callback)arm7_read_word_gen;
		arm7_rd32[i] = (mmu_read_dword_callback)arm7_read_dword_gen;
		arm7_wr8[i]  = (mmu_write_byte_callback)arm7_write_byte_gen;
		arm7_wr16[i] = (mmu_write_word_callback)arm7_write_word_gen;
		arm7_wr32[i] = (mmu_write_dword_callback)arm7_write_dword_gen;
	}
	for (i=0; i<16; i++) {
		arm9_rd8[0x80+i]  = (mmu_read_byte_callback)arm9_read_byte_rom;
		arm9_rd8[0x90+i]  = (mmu_read_byte_callback)arm9_read_byte_rom;
		arm9_rd8[0xA0+i]  = (mmu_read_byte_callback)arm9_read_byte_sram;
		arm9_rd8[0x60+i]  = (mmu_read_byte_callback)arm9_read_byte_vram;
		arm9_rd16[0x80+i] = (mmu_read_word_callback)arm9_read_word_rom;
		arm9_rd16[0x90+i] = (mmu_read_word_callback)arm9_read_word_rom;
		arm9_rd16[0xA0+i] = (mmu_read_word_callback)arm9_read_word_zero;
		arm9_rd16[0x60+i] = (mmu_read_word_callback)arm9_read_word_vram;
		arm9_rd32[0x80+i] = (mmu_read_dword_callback)arm9_read_dword_rom;
		arm9_rd32[0x90+i] = (mmu_read_dword_callback)arm9_read_dword_rom;
		arm9_rd32[0xA0+i] = (mmu_read_dword_callback)arm9_read_dword_zero;
		arm9_rd32[0x60+i] = (mmu_read_dword_callback)arm9_read_dword_vram;

		arm9_wr8[0x00+i]  = (mmu_write_byte_callback)arm9_write_byte_na;
		arm9_wr8[0x80+i]  = (mmu_write_byte_callback)arm9_write_byte_na;
		arm9_wr8[0x90+i]  = (mmu_write_byte_callback)arm9_write_byte_na;
		arm9_wr8[0xA0+i]  = (mmu_write_byte_callback)arm9_write_byte_sram;
		arm9_wr8[0xB0+i]  = (mmu_write_byte_callback)arm9_write_byte_na;
		arm9_wr8[0xC0+i]  = (mmu_write_byte_callback)arm9_write_byte_na;
		arm9_wr8[0xD0+i]  = (mmu_write_byte_callback)arm9_write_byte_na;
		arm9_wr8[0xE0+i]  = (mmu_write_byte_callback)arm9_write_byte_na;
		arm9_wr8[0xF0+i]  = (mmu_write_byte_callback)arm9_write_byte_na;
		arm9_wr8[0x60+i]  = (mmu_write_byte_callback)arm9_write_byte_vram;
		arm9_wr16[0x00+i] = (mmu_write_word_callback)arm9_write_word_na;
		arm9_wr16[0x80+i] = (mmu_write_word_callback)arm9_write_word_na;
		arm9_wr16[0x90+i] = (mmu_write_word_callback)arm9_write_word_na;
		arm9_wr16[0xA0+i] = (mmu_write_word_callback)arm9_write_word_na;
		arm9_wr16[0xB0+i] = (mmu_write_word_callback)arm9_write_word_na;
		arm9_wr16[0xC0+i] = (mmu_write_word_callback)arm9_write_word_na;
		arm9_wr16[0xD0+i] = (mmu_write_word_callback)arm9_write_word_na;
		arm9_wr16[0xE0+i] = (mmu_write_word_callback)arm9_write_word_na;
		arm9_wr16[0xF0+i] = (mmu_write_word_callback)arm9_write_word_na;
		arm9_wr16[0x60+i] = (mmu_write_word_callback)arm9_write_word_vram;
		arm9_wr32[0x00+i] = (mmu_write_dword_callback)arm9_write_dword_na;
		arm9_wr32[0x80+i] = (mmu_write_dword_callback)arm9_write_dword_na;
		arm9_wr32[0x90+i] = (mmu_write_dword_callback)arm9_write_dword_na;
		arm9_wr32[0xA0+i] = (mmu_write_dword_callback)arm9_write_dword_na;
		arm9_wr32[0xB0+i] = (mmu_write_dword_callback)arm9_write_dword_na;
		arm9_wr32[0xC0+i] = (mmu_write_dword_callback)arm9_write_dword_na;
		arm9_wr32[0xD0+i] = (mmu_write_dword_callback)arm9_write_dword_na;
		arm9_wr32[0xE0+i] = (mmu_write_dword_callback)arm9_write_dword_na;
		arm9_wr32[0xF0+i] = (mmu_write_dword_callback)arm9_write_dword_na;
		arm9_wr32[0x60+i] = (mmu_write_dword_callback)arm9_write_dword_vram;

		arm7_rd8[0x80+i]  = (mmu_read_byte_callback)arm7_read_byte_rom;
		arm7_rd8[0x90+i]  = (mmu_read_byte_callback)arm7_read_byte_rom;
		arm7_rd8[0xA0+i]  = (mmu_read_byte_callback)arm7_read_byte_sram;
		arm7_rd8[0x60+i]  = (mmu_read_byte_callback)arm7_read_byte_vram;
		arm7_rd16[0x80+i] = (mmu_read_word_callback)arm7_read_word_rom;
		arm7_rd16[0x90+i] = (mmu_read_word_callback)arm7_read_word_rom;
		arm7_rd16[0xA0+i] = (mmu_read_word_callback)arm7_read_word_zero;
		arm7_rd16[0x60+i] = (mmu_read_word_callback)arm7_read_word_vram;
		arm7_rd32[0x80+i] = (mmu_read_dword_callback)arm7_read_dword_rom;
		arm7_rd32[0x90+i] = (mmu_read_dword_callback)arm7_read_dword_rom;
		arm7_rd32[0xA0+i] = (mmu_read_dword_callback)arm7_read_dword_zero;
		arm7_rd32[0x60+i] = (mmu_read_dword_callback)arm7_read_dword_vram;

		arm7_wr8[0x00+i] = (mmu_write_byte_callback)arm7_write_byte_na;
		arm7_wr8[0x80+i] = (mmu_write_byte_callback)arm7_write_byte_na;
		arm7_wr8[0x90+i] = (mmu_write_byte_callback)arm7_write_byte_na;
		arm7_wr8[0xA0+i] = (mmu_write_byte_callback)arm7_write_byte_sram;
		arm7_wr8[0xB0+i] = (mmu_write_byte_callback)arm7_write_byte_na;
		arm7_wr8[0xC0+i] = (mmu_write_byte_callback)arm7_write_byte_na;
		arm7_wr8[0x60+i] = (mmu_write_byte_callback)arm7_write_byte_vram;
		arm7_wr16[0x00+i] = (mmu_write_word_callback)arm7_write_word_na;
		arm7_wr16[0x80+i] = (mmu_write_word_callback)arm7_write_word_na;
		arm7_wr16[0x90+i] = (mmu_write_word_callback)arm7_write_word_na;
		arm7_wr16[0xA0+i] = (mmu_write_word_callback)arm7_write_word_na;
		arm7_wr16[0xB0+i] = (mmu_write_word_callback)arm7_write_word_na;
		arm7_wr16[0xC0+i] = (mmu_write_word_callback)arm7_write_word_na;
		arm7_wr16[0x60+i] = (mmu_write_word_callback)arm7_write_word_vram;
		arm7_wr32[0x00+i] = (mmu_write_dword_callback)arm7_write_dword_na;
		arm7_wr32[0x80+i] = (mmu_write_dword_callback)arm7_write_dword_na;
		arm7_wr32[0x90+i] = (mmu_write_dword_callback)arm7_write_dword_na;
		arm7_wr32[0xA0+i] = (mmu_write_dword_callback)arm7_write_dword_na;
		arm7_wr32[0xB0+i] = (mmu_write_dword_callback)arm7_write_dword_na;
		arm7_wr32[0xC0+i] = (mmu_write_dword_callback)arm7_write_dword_na;
		arm7_wr32[0x60+i] = (mmu_write_dword_callback)arm7_write_dword_vram;

		arm9_rd8[0x30+i]  = (mmu_read_byte_callback)arm9_read_byte_wram;
		arm9_rd16[0x30+i] = (mmu_read_word_callback)arm9_read_word_wram;
		arm9_rd32[0x30+i] = (mmu_read_dword_callback)arm9_read_dword_wram;
		arm9_wr8[0x30+i]  = (mmu_write_byte_callback)arm9_write_byte_wram;
		arm9_wr16[0x30+i] = (mmu_write_word_callback)arm9_write_word_wram;
		arm9_wr32[0x30+i] = (mmu_write_dword_callback)arm9_write_dword_wram;

		arm7_rd16[0x30+(i&7)] = (mmu_read_word_callback)arm7_read_word_wram;
		arm7_rd32[0x30+(i&7)] = (mmu_read_dword_callback)arm7_read_dword_wram;
		arm7_wr16[0x30+(i&7)] = (mmu_write_word_callback)arm7_write_word_wram;
		arm7_wr32[0x30+(i&7)] = (mmu_write_dword_callback)arm7_write_dword_wram;

		arm7_rd8[0x38+(i&7)]  = (mmu_read_byte_callback)arm7_read_byte_ewram;
		arm7_rd16[0x38+(i&7)] = (mmu_read_word_callback)arm7_read_word_ewram;
		arm7_rd32[0x38+(i&7)] = (mmu_read_dword_callback)arm7_read_dword_ewram;
		arm7_wr8[0x38+(i&7)]  = (mmu_write_byte_callback)arm7_write_byte_ewram;
		arm7_wr16[0x38+(i&7)] = (mmu_write_word_callback)arm7_write_word_ewram;
		arm7_wr32[0x38+(i&7)] = (mmu_write_dword_callback)arm7_write_dword_ewram;

		arm7_rd8[0xF0+i]  = (mmu_read_byte_callback)arm7_read_byte_na;
		arm7_rd16[0xF0+i] = (mmu_read_word_callback)arm7_read_word_na;
		arm7_rd32[0xF0+i] = (mmu_read_dword_callback)arm7_read_dword_na;
		arm7_wr8[0xF0+i]  = (mmu_write_byte_callback)arm7_write_byte_na;
		arm7_wr16[0xF0+i] = (mmu_write_word_callback)arm7_write_word_na;
		arm7_wr32[0xF0+i] = (mmu_write_dword_callback)arm7_write_dword_na;
	}
	arm9_rd16[ipc_page] = (mmu_read_word_callback)arm9_read_word_ipc;

	rdh_ipc        = (mmu_read_word_callback)arm9_read_word_ipc;
	rdh_ipc_mirror = (mmu_read_word_callback)arm9_read_word_ipc_mirror;

	arm9_rd8[0x40]  = (mmu_read_byte_callback)arm9_read_byte_io;
	arm9_rd8[0x41]  = (mmu_read_byte_callback)arm9_read_byte_io;

	arm9_rd16[0x27] = arm9_rd16[0x23];
	arm9_rd16[0x40] = (mmu_read_word_callback)arm9_read_word_io;
	arm9_rd16[0x41] = (mmu_read_word_callback)arm9_read_word_io;
	arm9_rd16[0x50] = (mmu_read_word_callback)arm9_read_word_plt;
	arm9_rd16[0x68] = (mmu_read_word_callback)arm9_read_word_lcdc;
	arm9_rd16[0x70] = (mmu_read_word_callback)arm9_read_word_oam;

	arm9_rd32[0x40] = (mmu_read_dword_callback)arm9_read_dword_io;
	arm9_rd32[0x41] = (mmu_read_dword_callback)arm9_read_dword_io;
	arm9_rd32[0x68] = (mmu_read_dword_callback)arm9_read_dword_lcdc;

	arm9_wr8[0x40]  = (mmu_write_byte_callback)arm9_write_byte_io;

	arm9_wr16[0x40] = (mmu_write_word_callback)arm9_write_word_io;
	arm9_wr16[0x50] = (mmu_write_word_callback)arm9_write_word_plt;
	arm9_wr16[0x68] = (mmu_write_word_callback)arm9_write_word_lcdc;
	arm9_wr16[0x70] = (mmu_write_word_callback)arm9_write_word_oam;

	arm9_wr32[0x40] = (mmu_write_dword_callback)arm9_write_dword_io;
	arm9_wr32[0x50] = (mmu_write_dword_callback)arm9_write_dword_plt;
	arm9_wr32[0x68] = (mmu_write_dword_callback)arm9_write_dword_lcdc;
	arm9_wr32[0x70] = (mmu_write_dword_callback)arm9_write_dword_oam;

	arm7_rd8[0x40]  = (mmu_read_byte_callback)arm7_read_byte_io;
	arm7_rd8[0x41]  = (mmu_read_byte_callback)arm7_read_byte_io;

	arm7_rd16[0x40] = (mmu_read_word_callback)arm7_read_word_io;
	arm7_rd16[0x41] = (mmu_read_word_callback)arm7_read_word_io;
	arm7_rd16[0x50] = (mmu_read_word_callback)arm7_read_word_plt;
	arm7_rd16[0x68] = (mmu_read_word_callback)arm7_read_word_lcdc;
	arm7_rd16[0x70] = (mmu_read_word_callback)arm7_read_word_oam;
	arm7_rd32[0x40] = (mmu_read_dword_callback)arm7_read_dword_io;
	arm7_rd32[0x41] = (mmu_read_dword_callback)arm7_read_dword_io;
	arm7_rd32[0x68] = (mmu_read_dword_callback)arm7_read_dword_lcdc;
	arm7_wr8[0x40]  = (mmu_write_byte_callback)arm7_write_byte_io;
	arm7_wr16[0x40] = (mmu_write_word_callback)arm7_write_word_io;
	arm7_wr16[0x50] = (mmu_write_word_callback)arm7_write_word_plt;
	arm7_wr16[0x68] = (mmu_write_word_callback)arm7_write_word_lcdc;
	arm7_wr16[0x70] = (mmu_write_word_callback)arm7_write_word_oam;
	arm7_wr32[0x40] = (mmu_write_dword_callback)arm7_write_dword_io;
	arm7_wr32[0x50] = (mmu_write_dword_callback)arm7_write_dword_plt;
	arm7_wr32[0x68] = (mmu_write_dword_callback)arm7_write_dword_lcdc;
	arm7_wr32[0x70] = (mmu_write_dword_callback)arm7_write_dword_oam;

	hout = GetStdHandle(STD_OUTPUT_HANDLE);
	return true;

}


// Align upwards to the nearest power of two
int align_pow2(int n) {
	dummy = n;
	__asm {
		pusha
		mov eax,[dummy]
		bsr ecx,eax
		inc ecx
		mov eax,1
		shl eax,cl
		dec eax
		mov [dummy],eax
		popa
	}
	return dummy;
}





bool __stdcall mmu_reset() {
	int			i;
	static int	dwNull = 0;
	static int	ask = ASK;
	unsigned char *GBACart;

	zombie(TELL,ZOMBIE_DEBUG_MSG,"MMU reset",(void*)dwNull);

	cpu_irq   = (cpu_irq_callback)zombie(ask,CPU_IRQ_CALLBACK,(void*)dwNull,(void*)dwNull);
	arm7_irq  = (cpu_irq_callback)zombie(ask,CPU_IRQ_CALLBACK,(void*)dwNull,(void*)1);
	arm9_dma  = (cpu_dma_callback)zombie(ask,CPU_DMA_CALLBACK,(void*)dwNull,(void*)dwNull);
	arm7_dma  = (cpu_dma_callback)zombie(ask,CPU_DMA_CALLBACK,(void*)dwNull,(void*)1);
	arm9_regs = (int*)zombie(ask,CPU_REG,(void*)dwNull,(void*)dwNull);
	arm7_regs = (int*)zombie(ask,CPU_REG,(void*)dwNull,(void*)1);
	gpu_set   = (gpu_set_callback)zombie(ask,GPU_SET_CALLBACK,(void*)dwNull,(void*)dwNull);
	gpu_get   = (gpu_get_callback)zombie(ask,GPU_GET_CALLBACK,(void*)dwNull,(void*)dwNull);
	apu_write = (apu_write_callback)zombie(ask,APU_WRITE_CALLBACK,(void*)dwNull,(void*)dwNull);

	touch     = (touchpad*)zombie(ask,TOUCHPAD_DATA,(void*)dwNull,(void*)dwNull);

	//cflash_init();

	cart      = (unsigned char*)zombie(ask,ZOMBIE_ROM_BASE,(void*)dwNull,(void*)dwNull);
	if ((int)cart == dwNull) {
		zombie(TELL,ZOMBIE_DEBUG_MSG,"MMU: bad ROM handle",(void*)dwNull);
		return false;
	}
	ROM = cart;

	ROMsub = (unsigned char*)zombie(ask,ZOMBIE_ROM_BASE,(void*)1,(void*)dwNull);

	VRAM = (unsigned char*)zombie(ask,GPU_MEMORY,(void*)dwNull,(void*)dwNull);
	if ((int)VRAM == dwNull)
		return false;
	for (i=0; i<9; i++)
		vramBank[i].type = DISABLED;
	for (i=0; i<1024; i++)
		vramChunk[i] = NULL;

	VRAM_A		= VRAM;
	VRAM_B		= VRAM_A + 0x20000;
	VRAM_C		= VRAM_B + 0x20000;
	VRAM_D		= VRAM_C + 0x20000;
	VRAM_E		= VRAM_D + 0x20000;
	VRAM_F		= VRAM_E + 0x10000;
	VRAM_G		= VRAM_F + 0x4000;
	VRAM_H		= VRAM_G + 0x4000;
	VRAM_I		= VRAM_H + 0x8000;
	vramBase[0] = VRAM_A;
	vramBase[1] = VRAM_B;
	vramBase[2] = VRAM_C;
	vramBase[3] = VRAM_D;
	vramBase[4] = VRAM_E;
	vramBase[5] = VRAM_F;
	vramBase[6] = VRAM_G;
	vramBase[7] = VRAM_H;
	vramBase[8] = VRAM_I;
	
	//remap_vram(2,0x84);
	//remap_vram(7,0x81);

	systime = (SYSTEMTIME*)zombie(ask,ZOMBIE_TIME,(void*)dwNull,(void*)dwNull);


	//zombie(TELL,ZOMBIE_DEBUG_MSG,"Setting reset vector to 0x02004000",0);

	IOREGS   = &VRAM[IOREGS_OFFSET];
	IOREGS_ARM7   = &VRAM[IOREGS_ARM7_OFFSET];
	PLT_VRAM = &VRAM[PALETTE_OFFSET];
	OAM      = &VRAM[SPRITE_TABLE_OFFSET];

	IOREGS[0x130] = 0xFF;
	keys          = 0x3FF;
	btnXY[0]      = btnXY[1] = 0;

	for (i=0; i<0x10; i++) {
		arm9_pageTable[0x30+i] = IWRAM;
		arm9_pageTable[0x20+i] = EXRAM + ((i&3)*0x100000);
		arm9_pageTable[0x00+i] = DTCM; 
		arm9_pageTable[0xF0+i] = SYSROM;

		arm7_pageTable[0x00+i]     = SYSROM_ARM7;
		arm7_pageTable[0x30+(i&7)] = EIWRAM;
		arm7_pageTable[0x38+(i&7)] = EIWRAM;
		arm7_pageTable[0x20+i]     = EXRAM + ((i&3)*0x100000);

		arm7_rd16[0x30+(i&7)] = (mmu_read_word_callback)arm7_read_word_ewram;
		arm7_rd32[0x30+(i&7)] = (mmu_read_dword_callback)arm7_read_dword_ewram;
		arm7_wr16[0x30+(i&7)] = (mmu_write_word_callback)arm7_write_word_ewram;
		arm7_wr32[0x30+(i&7)] = (mmu_write_dword_callback)arm7_write_dword_ewram;		
	}

	arm9_pageLength[8] = (unsigned int)zombie(ask,ZOMBIE_ROM_SIZE,(void*)dwNull,(void*)dwNull);
	arm9_pageLength[9] = 0;
	arm9_pageLength[7] = 0x800;
	arm9_pageLength[6] = 0x100000;
	arm9_pageLength[5] = 0x1000;
	arm9_pageLength[4] = 0x2000;

	arm7_pageLength[8] = (unsigned int)zombie(ask,ZOMBIE_ROM_SIZE,(void*)1,(void*)dwNull);
	arm7_pageLength[9] = 0;
	arm7_pageLength[7] = 0x800;
	arm7_pageLength[6] = 0x100000;
	arm7_pageLength[5] = 0x1000;
	arm7_pageLength[4] = 0x2000;

	GBACart = (unsigned char*)zombie(ask,ZOMBIE_GBA_CART,(void*)dwNull,(void*)dwNull);
	for (i=0; i<0x10; i++) {
		if (GBACart) {
			arm9_pageTable[0x80+i] = GBACart + i*0x100000;
			if (arm9_pageLength[8] > 0x01000000)
				arm9_pageTable[0x90+i] = GBACart + 0x1000000 + i*0x100000;
			else
				arm9_pageTable[0x90+i] = GBACart + i*0x100000;

			//arm9_pageTable[0xA0+i] = GBACart + i*0x100000;
			//arm9_pageTable[0xC0+i] = GBACart + i*0x100000;

			arm7_pageTable[0x80+i] = GBACart + i*0x100000;
			if (arm9_pageLength[8] > 0x01000000)
				arm7_pageTable[0x90+i] = GBACart + 0x1000000 + i*0x100000;
			else
				arm7_pageTable[0x90+i] = GBACart + i*0x100000;
			//arm7_pageTable[0xA0+i] = GBACart + i*0x100000;
			//arm7_pageTable[0xC0+i] = GBACart + i*0x100000;
		} else {
			arm9_pageTable[0x80+i] = ROM + i*0x100000;
			arm9_pageTable[0x90+i] = ROM + i*0x100000;
			//arm9_pageTable[0xA0+i] = ROM + i*0x100000;
			//arm9_pageTable[0xC0+i] = ROM + i*0x100000;

			arm7_pageTable[0x80+i] = ROMsub + i*0x100000;
			arm7_pageTable[0x90+i] = ROMsub + i*0x100000;
			//arm7_pageTable[0xA0+i] = ROMsub + i*0x100000;
			//arm7_pageTable[0xC0+i] = ROMsub + i*0x100000;
		}
	}
	arm9_pageTable[0x70] = OAM; 
	arm9_pageTable[0x60] = VRAM;
	arm9_pageTable[0x62] = S_VRAM; 
	arm9_pageTable[0x64] = OBJ_VRAM;
	arm9_pageTable[0x66] = OBJ_S_VRAM;  
	arm9_pageTable[0x68] = LCDC_VRAM;
	arm9_pageTable[0x50] = PLT_VRAM; 
	arm9_pageTable[0x40] = IOREGS;
	arm9_pageTable[0x41] = IOREGS;


	arm7_pageTable[0x70] = OAM; 
	arm7_pageTable[0x60] = VRAM;
	arm7_pageTable[0x62] = S_VRAM; 
	arm7_pageTable[0x64] = OBJ_VRAM;
	arm7_pageTable[0x66] = OBJ_S_VRAM;  
	arm7_pageTable[0x68] = LCDC_VRAM;
	arm7_pageTable[0x50] = PLT_VRAM; 
	arm7_pageTable[0x40] = IOREGS_ARM7;
	arm7_pageTable[0x41] = IOREGS_ARM7;

	if (arm9_pageLength[8] > 0x01000000) {
		arm9_pageLength[9] = arm9_pageLength[8] - 0x01000000;
		arm9_pageLength[8] = 0x01000000;
		arm7_pageLength[9] = arm7_pageLength[8] - 0x01000000;
		arm7_pageLength[8] = 0x01000000;
	}

	dtcmPage = 0x08;
	dtcmLast.nPage = -1;
	arm9_pageTable[dtcmPage] = DTCM;
	arm9_pageLength[dtcmPage>>4] = 0x4000;

	arm9_rd8[dtcmPage]  = (mmu_read_byte_callback)arm9_read_byte_dtcm;
	arm9_rd16[dtcmPage] = (mmu_read_word_callback)arm9_read_word_dtcm;
	arm9_rd32[dtcmPage] = (mmu_read_dword_callback)arm9_read_dword_dtcm;
	arm9_wr8[dtcmPage]  = (mmu_write_byte_callback)arm9_write_byte_dtcm;
	arm9_wr16[dtcmPage] = (mmu_write_word_callback)arm9_write_word_dtcm;
	arm9_wr32[dtcmPage] = (mmu_write_dword_callback)arm9_write_dword_dtcm;

	arm9_rd8[0]  = (mmu_read_byte_callback)arm9_read_byte_itcm;
	arm9_rd16[0] = (mmu_read_word_callback)arm9_read_word_itcm;
	arm9_rd32[0] = (mmu_read_dword_callback)arm9_read_dword_itcm;
	arm9_wr8[0]  = (mmu_write_byte_callback)arm9_write_byte_itcm;
	arm9_wr16[0] = (mmu_write_word_callback)arm9_write_word_itcm;
	arm9_wr32[0] = (mmu_write_dword_callback)arm9_write_dword_itcm;

	/* Set up addresses for GBAMP */
	CF_REG_DATA = 0x9000000;
	CF_REG_ERR	= 0x9020000;
	CF_REG_SEC	= 0x9040000;
	CF_REG_LBA1	= 0x9060000;
	CF_REG_LBA2	= 0x9080000;
	CF_REG_LBA3	= 0x90A0000;
	CF_REG_LBA4	= 0x90C0000;
	CF_REG_CMD	= 0x90E0000;
	CF_REG_STS	= 0x98C0000;

	for (i=0; i<9; i++) {
		arm9_rd16[0x90+i] = (mmu_read_word_callback)arm9_read_word_rom;
		arm9_rd8[0x90+i] = (mmu_read_byte_callback)arm9_read_byte_rom;
		arm9_wr16[0x90+i] = (mmu_write_word_callback)arm9_write_word_na;
	}
	cflash_close();
	if (cfDevice==1) {
		cflash_init();
		if (cflashDeviceEnabled) {
			for (i=0; i<9; i++) {
				arm9_rd16[0x90+i] = (mmu_read_word_callback)arm9_read_word_cf;
				arm9_rd8[0x90+i]  = (mmu_read_byte_callback)arm9_read_byte_cf;
				arm9_wr16[0x90+i] = (mmu_write_word_callback)arm9_write_word_cf;
			}
		}
	}

	for (i=0; i<256; i++) {
		arm9_pageMask[i] = arm9_pageLength[i>>4] - 1;
		arm7_pageMask[i] = arm7_pageLength[i>>4] - 1;
	}

	// Align the ROM size masks to the nearest 2^n boundary
	for (i=0; i<16; i++) {
		arm7_pageMask[0x38+(i&7)] = 0xFFFF;
		arm9_pageMask[0x80+i] = align_pow2(arm9_pageLength[8]);
		arm7_pageMask[0x80+i] = align_pow2(arm7_pageLength[8]);
		arm9_pageMask[0x90+i] = align_pow2(arm9_pageLength[9]);
		arm7_pageMask[0x90+i] = align_pow2(arm7_pageLength[9]);
	}

	IOREGS[0x247] = 3;
	IOREGS_ARM7[0x241] = 3;
	remap_wram(0);
	remap_wram(1);

	// Initialize timers and DMA channels
	for (i=0; i<4; i++)	{
		timer[0][i].enabled = false;
		timer[0][i].countup = false;
		timer[0][i].tcurr = 0;
		timer[0][i].tstart = 0;

		timer[1][i].enabled = false;
		timer[1][i].countup = false;
		timer[1][i].tcurr = 0;
		timer[1][i].tstart = 0;

		dma[0][i].sad_int = 0;
		dma[0][i].dad_int = 0;
		dma[0][i].enabled = false;
		dma[1][i].sad_int = 0;
		dma[1][i].dad_int = 0;
		dma[1][i].enabled = false;
	}

	// Clear IWRAM
	//memset(IWRAM,0,iwramSize);

	if (useExtBios)	{
		memcpy(SYSROM,BIOS,0x4000);
	} else {
		memcpy(SYSROM,              sysromcode_0,       8*4);
		memcpy(SYSROM_ARM7,         sysromcode_0,       8*4);
		memcpy(SYSROM+0x128,        sysromcode_128,     9*4);
		memcpy(SYSROM_ARM7+0x128,   arm7_sysromcode_128,6*4);

		memcpy(SYSROM+0xF0000,      sysromcode_0,       8*4);
		memcpy(SYSROM+0xF0000+0x128,sysromcode_128,     9*4);
	}

	*(arm9_pageTable[0x23] + 0x7F016) = 5;
	*(arm9_pageTable[0x23] + 0x7F017) = 6;	// y
	*(arm9_pageTable[0x23] + 0x7F018) = 7;	// m
	*(arm9_pageTable[0x23] + 0x7F019) = 8;	// d
	*(arm9_pageTable[0x23] + 0x7F01A) = 9;
	*(arm9_pageTable[0x23] + 0x7F01B) = 1;	// h
	*(arm9_pageTable[0x23] + 0x7F01C) = 2;	// m
	*(arm9_pageTable[0x23] + 0x7F01D) = 3;	// s

	srand(123);

	arm9RecvFifoPos = arm9SendFifoPos = 0;
	arm7RecvFifoPos = arm7SendFifoPos = 0;
	arm9FifoCnt = arm7FifoCnt = 0x101;
	IOREGS[0x180] = IOREGS_ARM7[0x180] = 0;
	IOREGS[0x181] = IOREGS_ARM7[0x181] = 0;
	IOREGS[0x184] = IOREGS_ARM7[0x184] = 1;	// Send FIFO empty
	IOREGS[0x185] = IOREGS_ARM7[0x185] = 1;	// Recv FIFO empty
	IOREGS[0x188] = IOREGS_ARM7[0x188] = 0;
	IOREGS[0x189] = IOREGS_ARM7[0x189] = 0;

	ipcSync = ipcSync7 = 0;

	irqFlags = 0;
	irqFlags7 = 0;
	irqAck = 0xFFFF;
	*(int*)(IOREGS + IF_) = 0;
	*(short int*)(arm9_pageTable[0x08] + 0x3FF8) = 0; 
	*(short int*)(arm7_pageTable[0x38] + 0xFFF8) = 0; 

	return true;
}


void __stdcall mmu_close() {
	cflash_close();
	LocalFree(mmu_memory);
}



int __stdcall mmu_config(HWND hwnd) {
	return DialogBoxParam(hInst,
            MAKEINTRESOURCE(IDD_DIALOG1), hwnd,
            DlgProc, NULL);
}




unsigned char* __stdcall mmu_translate_address(unsigned int address) {
	return arm9_pageTable[address>>20] + (address&0x0FFFFF);
}

unsigned char* __stdcall arm7_translate_address(unsigned int address) {
	return arm7_pageTable[address>>20] + (address&0x0FFFFF);
}



void dma_transfer(void *src,void *dest,int cnt) {
	memcpy(dest,src,cnt);
}



unsigned short int __stdcall mmu_read_word(unsigned int);
void WRHDECL mmu_write_word(unsigned int,int);
int __stdcall mmu_read_dword(unsigned int);
void WRWDECL mmu_write_dword(unsigned int,int);
void __stdcall mmu_irq(int);


void dma_setup(int proc, int channel, bool forceDMA) {
	int						i;
	unsigned int			cnt;
	unsigned char			sadcnt,dadcnt;
	unsigned short int		cnts,cnth;
	unsigned int			src,dest,srcend,destend;
	int						srcinc,dstinc;
	unsigned char			*io_base;
	unsigned int			*cycptr;
	unsigned int			srcadr,dstadr;
	mmu_read_word_callback	*rdh;
	mmu_write_word_callback *wrh;
	mmu_read_dword_callback *rdw;
	mmu_write_dword_callback *wrw;


	io_base = (proc==0)?IOREGS:IOREGS_ARM7;
	cycptr  = (proc==0)?arm9_cycle:arm7_cycle;

	if ( ((io_base[dma[proc][channel].cnt_h+1] & 0xB0) == 0x80) ||
		 (((io_base[dma[proc][channel].cnt_h+1] & 0xB0) >= 0x80) && forceDMA) )
	{
		if (dma[proc][channel].enabled == false) {
			cnts = *(unsigned short*)(io_base + dma[proc][channel].cnt_l);
			cnt = cnts;
			cnt += (io_base[dma[proc][channel].cnt_h]&0x1F)<<16;
			if ((cnt == 0) || (cnt > dmamax[channel]))
				cnt = dmamax[channel];
			dma[proc][channel].cnt_int = cnt;
		}
		cnt = dma[proc][channel].cnt_int;
		cnt <<= ((io_base[dma[proc][channel].cnt_h+1] & 4) == 4) ? 2 : 1;

		src = (*(unsigned int*)(io_base + dma[proc][channel].sad)) & 0xFFFFFFF;
		dest = (*(unsigned int*)(io_base + dma[proc][channel].dad)) & 0xFFFFFFF;
		if (dma[proc][channel].enabled == false) {
			dma[proc][channel].sad_int = src;
			dma[proc][channel].dad_int = dest;
		}
		dma[proc][channel].enabled = true;

		sadcnt = io_base[dma[proc][channel].cnt_h] & 0x80;
		sadcnt |= io_base[dma[proc][channel].cnt_h+1] & 0x01;
		dadcnt = (io_base[dma[proc][channel].cnt_h]>>5)&3;

		cnth = io_base[dma[proc][channel].cnt_h] + (io_base[dma[proc][channel].cnt_h+1]<<8);

		/*if ((dma[proc][channel].dad_int >= 0x6800000)&&
		(dma[proc][channel].dad_int < 0x7000000)) {
			if (0) {*/
		if (logDMA) {
			wsprintf(buffer,"ARM%d DMA%d %d bytes from %08x to %08x, CNT = %08x, PC=%08X\n",
				9-proc*2,channel,cnt,dma[proc][channel].sad_int,dma[proc][channel].dad_int,
				*(int*)&IOREGS[dma[proc][channel].cnt_l],(proc==0)?arm9_regs[15]:arm7_regs[15]);
			//WriteConsole(hout,buffer,lstrlen(buffer),(unsigned long*)&dummy,NULL);
			zombie(TELL,ZOMBIE_DEBUG_MSG,buffer,(void*)0L);
		}
		/*}
		}*/

		if (sadcnt == 0x00)	{
			/*srcend = (dma[channel].sad_int + cnt) & 0xFFFFFF;
			if (srcend >= pageLength[dma[channel].sad_int >> 24])
				cnt -= (srcend - pageLength[dma[channel].sad_int >> 24]);
			if (cnt<=0) return;*/
		}


		//dma_transfer((void*)mmu_translate_address(src),(void*)mmu_translate_address(dest),cnt);

		if (blockDMA == 0) {
		if ((io_base[dma[proc][channel].cnt_h+1] & 4) == 4) {
			// Increment source
			if (sadcnt == 0x00)	{
				if (dadcnt==2) {
					if (proc==0)
						for (i=0; i<cnt; i+=4) {
							arm9_wr32[((dma[proc][channel].dad_int)>>20)&0xFF](dma[proc][channel].dad_int,
								arm9_rd32[((dma[proc][channel].sad_int+i)>>20)&0xFF](dma[proc][channel].sad_int+i));
						}
					else
						for (i=0; i<cnt; i+=4) {
							arm7_wr32[((dma[proc][channel].dad_int)>>20)&0xFF](dma[proc][channel].dad_int,
								arm7_rd32[((dma[proc][channel].sad_int+i)>>20)&0xFF](dma[proc][channel].sad_int+i));
						}
					*cycptr += cnt/2;
				} else {
					if (proc==0) {
						if ((int)arm9_dma == 0)	{
							for (i=0; i<cnt; i+=4)
									mmu_write_dword(dma[proc][channel].dad_int+i,mmu_read_dword(dma[proc][channel].sad_int+i));
						} else {
							arm9_dma((char*)dma[proc][channel].dad_int,(char*)dma[proc][channel].sad_int,cnt,4,4);
						}
					} else {
						if ((int)arm7_dma!=0) 
							arm7_dma((char*)dma[proc][channel].dad_int,(char*)dma[proc][channel].sad_int,cnt,4,4);
					}
				}

			// Fixed source
			} else if (sadcnt == 0x01) {
				/*for (i=0; i<cnt; i+=4)
					arm9_wr32[((dma[proc][channel].dad_int+i)>>20)&0xFF](dma[proc][channel].dad_int+i,
						arm9_rd32[((dma[proc][channel].sad_int)>>20)&0xFF](dma[proc][channel].sad_int));
				*cycptr += cnt/2;*/
				rdw = (proc==0)?(&arm9_rd32[0]):(&arm7_rd32[0]);
				wrw = (proc==0)?(&arm9_wr32[0]):(&arm7_wr32[0]);
				if (dadcnt==1)
					dstinc = 4;
				else if (dadcnt==2)
					dstinc = 0;
				else
					dstinc = -4;
				srcinc = 0; //-4;
				srcadr = dma[proc][channel].sad_int;
				dstadr = dma[proc][channel].dad_int;
				for (i=0; i<cnt; i+=4) {
					wrw[dstadr>>20](dstadr,rdw[srcadr>>20](srcadr));
					//srcadr += srcinc;
					dstadr += dstinc;
				}
				*cycptr += cnt/2;

	
			} else if (sadcnt == 0x80) {
				rdw = (proc==0)?(&arm9_rd32[0]):(&arm7_rd32[0]);
				wrw = (proc==0)?(&arm9_wr32[0]):(&arm7_wr32[0]);
				if (dadcnt==1)
					dstinc = 4;
				else if (dadcnt==2)
					dstinc = 0;
				else
					dstinc = -4;
				srcinc = -4;
				srcadr = dma[proc][channel].sad_int;
				dstadr = dma[proc][channel].dad_int;
				for (i=0; i<cnt; i+=4) {
					wrw[dstadr>>20](dstadr,rdw[srcadr>>20](srcadr));
					srcadr += srcinc;
					dstadr += dstinc;
				}
				*cycptr += cnt/2;
			}
		} else {
			if (sadcnt == 0x00)	{
				if (dadcnt==2) {
					if (proc==0) {
						for (i=0; i<cnt; i+=2)
							arm9_wr16[((dma[proc][channel].dad_int)>>20)&0xFF](dma[proc][channel].dad_int,
										   arm9_rd16[((dma[proc][channel].sad_int+i)>>20)&0xFF](dma[proc][channel].sad_int+i));
					} else {
						for (i=0; i<cnt; i+=2)
							arm7_wr16[((dma[proc][channel].dad_int)>>20)&0xFF](dma[proc][channel].dad_int,
										   arm7_rd16[((dma[proc][channel].sad_int+i)>>20)&0xFF](dma[proc][channel].sad_int+i));
					}
					*cycptr += cnt/2;
				} else {
					if (proc==0) {
						if ((int)arm9_dma == 0)	{
							for (i=0; i<cnt; i+=2)
								arm9_wr16[((dma[proc][channel].dad_int+i)>>20)&0xFF](dma[proc][channel].dad_int+i,
											   arm9_rd16[((dma[proc][channel].sad_int+i)>>20)&0xFF](dma[proc][channel].sad_int+i));
						} else {
							arm9_dma((char*)dma[proc][channel].dad_int,(char*)dma[proc][channel].sad_int,cnt,2,2);
						}
					} else {
						if ((int)arm7_dma == 0)	{
							for (i=0; i<cnt; i+=2)
								arm7_wr16[((dma[proc][channel].dad_int+i)>>20)&0xFF](dma[proc][channel].dad_int+i,
											   arm7_rd16[((dma[proc][channel].sad_int+i)>>20)&0xFF](dma[proc][channel].sad_int+i));
						} else {
							arm7_dma((char*)dma[proc][channel].dad_int,(char*)dma[proc][channel].sad_int,cnt,2,2);
						}
					}
				}

			} else if (sadcnt == 0x01) {
				/*if (proc==0) {
					for (i=0; i<cnt; i+=2)
						arm9_wr16[((dma[proc][channel].dad_int+i)>>20)&0xFF](dma[proc][channel].dad_int+i,
						               arm9_rd16[((dma[proc][channel].sad_int)>>20)&0xFF](dma[proc][channel].sad_int));
					*cycptr += cnt/2;
				}*/
				rdh = (proc==0)?(&arm9_rd16[0]):(&arm7_rd16[0]);
				wrh = (proc==0)?(&arm9_wr16[0]):(&arm7_wr16[0]);
				if (dadcnt==1)
					dstinc = 2;
				else if (dadcnt==2)
					dstinc = 0;
				else
					dstinc = -2;
				srcinc = 0;
				srcadr = dma[proc][channel].sad_int;
				dstadr = dma[proc][channel].dad_int;
				for (i=0; i<cnt; i+=2) {
					wrh[dstadr>>20](dstadr,rdh[srcadr>>20](srcadr));
					//srcadr += srcinc;
					dstadr += dstinc;
				}
				*cycptr += cnt/2;

			} else if (sadcnt == 0x80) {
				rdh = (proc==0)?(&arm9_rd16[0]):(&arm7_rd16[0]);
				wrh = (proc==0)?(&arm9_wr16[0]):(&arm7_wr16[0]);
				if (dadcnt==1)
					dstinc = 2;
				else if (dadcnt==2)
					dstinc = 0;
				else
					dstinc = -2;
				srcinc = -2;
				srcadr = dma[proc][channel].sad_int;
				dstadr = dma[proc][channel].dad_int;
				for (i=0; i<cnt; i+=2) {
					wrh[dstadr>>20](dstadr,rdh[srcadr>>20](srcadr));
					srcadr += srcinc;
					dstadr += dstinc;
				}
				*cycptr += cnt/2;
			}

		}
		}

		switch (io_base[dma[proc][channel].cnt_h] & 0x60) {
			case 0x00:
				dma[proc][channel].dad_int += cnt;
				break;
			case 0x20:
				dma[proc][channel].dad_int -= cnt;
				break;
			default:
				break;
		}

		switch (sadcnt)	{
			case 0x00:
				dma[proc][channel].sad_int += cnt;
				break;
			case 0x80:
				dma[proc][channel].sad_int -= cnt;
				break;
			default:
				break;
		}

		// Reset the DMA enable bit unless repeat mode is specified
		if ((io_base[dma[proc][channel].cnt_h+1] & 0x02) == 0)	{
			io_base[dma[proc][channel].cnt_h+1] &= 0x7F;
			dma[proc][channel].enabled = false;
		}

		// Generate an IRQ if needed
		if (io_base[dma[proc][channel].cnt_h+1] & 0x40)
			if (io_base[IE+1]&(1<<channel))
				mmu_irq(0x100<<channel);
	} else
		dma[proc][channel].enabled = false;
}



void timer_setup(int proc, int channel) {
	static int deblimit = 20;
	unsigned char *iobase;
	unsigned int *cycptr,cntval;

	cycptr = (proc==0)?arm9_cycle:arm7_cycle;
	iobase = (proc==0)?IOREGS:IOREGS_ARM7;

	cntval = *(unsigned int*)(&iobase[TM0D+channel*4]);
	timer[proc][channel].unchanged = (timer[proc][channel].cntval == cntval)?true:false;
	timer[proc][channel].cntval = cntval;

	timer[proc][channel].tstart = (*cycptr)<<proc;
	timer[proc][channel].countup = ((iobase[TM0CNT+channel*4]&0x4)!=0);
	timer[proc][channel].td = *(unsigned short int*)&iobase[TM0D + channel*4];

	if (((iobase[TM0CNT+channel*4]&0x80)!=0)&&(timer[proc][channel].enabled==false))	{
		timer[proc][channel].tcurr = timer[proc][channel].td;
		timer[proc][channel].te = timer[proc][channel].td;
	}

	//timer[channel].enabled = false;
	if ((iobase[TM0CNT+channel*4]&0x40) &&
		(iobase[TM0CNT+channel*4]&0x80) &&
		(!timer[proc][channel].enabled) &&
		((iobase[TM0CNT+channel*4]&0x04)==0)) {
			/*if (proc==1) {
				wsprintf(buffer,"Adding overflow on channel %d at %X, %X\n",channel,
					(((0x10000-timer[proc][channel].tcurr)<<(prescalar[iobase[TM0CNT+channel*4]&3]))<<1),
					timer[proc][channel].tcurr);
				WriteConsole(hout,buffer,strlen(buffer),(LPDWORD)dummy,NULL);
			}*/
			zombie(TELL,ZOMBIE_EVENT,(void*)(EV_TIMER_OVERFLOW|((channel+proc*4)<<16)),
				   (void*)( timer[proc][channel].tstart + (((0x10000-timer[proc][channel].tcurr)<<(prescalar[iobase[TM0CNT+channel*4]&3]+1))) ) );
	
	} else if ((channel<3) && ((iobase[TM0CNT+channel*4]&0x80)!=0)) {
		if (timer[proc][channel+1].countup) {
			//wsprintf(buffer,"Adding event %d, %d, %X\n",proc,channel,(((0x10000-timer[proc][channel].tcurr)<<prescalar[iobase[TM0CNT+channel*4]&3]+1)));
			//WriteConsole(hout,buffer,strlen(buffer),(LPDWORD)dummy,NULL);
			if (iobase[TM0CNT+channel*4]&0x40) {
				zombie(TELL,ZOMBIE_EVENT,(void*)(EV_TIMER_OVERFLOW|((channel+proc*4)<<16)),
					   (void*)(timer[proc][channel].tstart + (((0x10000-timer[proc][channel].td)<<(prescalar[iobase[TM0CNT+channel*4]&3]+1))) ) );
			} else {
				timer[proc][channel].interval = (((0x10000-timer[proc][channel].td)<<(prescalar[iobase[TM0CNT+channel*4]&3]+1)));
				zombie(TELL,ZOMBIE_EVENT,(void*)(EV_TIMER_COUNTUP|((channel+proc*4)<<16)),
					   (void*)(timer[proc][channel].tstart + timer[proc][channel].interval ) );
			}
		}
	}
	if ((channel!=0) && ((iobase[TM0CNT+channel*4]&0x04)!=0)) {
		if (timer[proc][channel-1].enabled) {
			//wsprintf(buffer,"Starting countup timer %d, %d\n",proc,channel);
			//WriteConsole(hout,buffer,strlen(buffer),(LPDWORD)dummy,NULL);
			zombie(TELL,ZOMBIE_EVENT,(void*)(EV_TIMER_OVERFLOW|(((channel-1)+proc*4)<<16)),
			   (void*)(timer[proc][channel-1].tstart + ((0x10000-timer[proc][channel-1].td)<<(prescalar[iobase[TM0CNT+(channel-1)*4]&3]+1))));
		}
	}
	if (iobase[TM0CNT+channel*4]&0x80)
		timer[proc][channel].enabled = true;
	else
		timer[proc][channel].enabled = false;
}




void __stdcall mmu_irq(int mask) {
	int i;

	if (mask & 2) {
		if ((IOREGS[0x210]&2)&&(IOREGS[0x208]&1))
			irqFlags |= 2;
		if (IOREGS_ARM7[0x210]&2)
			irqFlags7 |= 2;
		for (i=1; i<4; i++) {
			if ((IOREGS_ARM7[dma[0][i].cnt_h+1] & 0xB0) == 0xA0)
				dma_setup(1,i,true);
			if ((IOREGS[dma[0][i].cnt_h+1] & 0xB0) == 0xA0)
				dma_setup(0,i,true);
		}
	}

	if (mask & 1) {
		//if (IOREGS[0x210]&1)
			irqFlags |= 1;
			irqFlags7 |= 1;
		for (i=1; i<4; i++) {
			if ((IOREGS_ARM7[dma[0][i].cnt_h+1] & 0xB0) == 0x90)
				dma_setup(1,i,true);
			if ((IOREGS[dma[0][i].cnt_h+1] & 0xB0) == 0x90)
				dma_setup(0,i,true);
		}
	}

	if (mask & 0x08) {
		irqFlags |= 0x08;
		if ((IOREGS[TM0CNT]&0x4)==0) {
			timer[0][0].enabled = false;
			timer_setup(0,0);
		}
		cpu_irq(0x0000018);
	}
	if (mask & 0x10) {
		//WriteConsole(hout,"IRQ on TM1\n",12,(LPDWORD)dummy,NULL);
		irqFlags |= 0x10;
		if ((IOREGS[TM1CNT]&0x4)==0) {
			timer[0][1].enabled = false;
			timer_setup(0,1);
		}
		cpu_irq(0x0000018);
	}
	if (mask & 0x20) {
		irqFlags |= 0x20;
		if ((IOREGS[TM2CNT]&0x4)==0) {
			timer[0][2].enabled = false;
			timer_setup(0,2);
		}
		cpu_irq(0x0000018);
	}
	if (mask & 0x40) {
		irqFlags |= 0x40;
		if ((IOREGS[TM3CNT]&0x4)==0) {
			timer[0][3].enabled = false;
			timer_setup(0,3);
		}
		cpu_irq(0x0000018);
	}

	// DMA
	if (mask & 0x400) {
		irqFlags |= 0x400;
		cpu_irq(0x0000018);
	}
	if (mask & 0x800) {
		irqFlags |= 0x800;
		cpu_irq(0x0000018);
	}


	if (mask & 0x080000) {
		irqFlags7 |= 0x08;
		if ((IOREGS_ARM7[TM0CNT]&0x4)==0) {
			timer[1][0].enabled = false;
			timer_setup(1,0);
		}
		arm7_irq(0x0000018);
	}
	if (mask & 0x100000) {
		irqFlags7 |= 0x10;
		if ((IOREGS_ARM7[TM1CNT]&0x4)==0) {
			timer[1][1].enabled = false;
			timer_setup(1,1);
		}
		arm7_irq(0x0000018);
	}
	if (mask & 0x200000) {
		irqFlags7 |= 0x20;
		if ((IOREGS_ARM7[TM2CNT]&0x4)==0) {
			timer[1][2].enabled = false;
			timer_setup(1,2);
		}
		arm7_irq(0x0000018);
	}
	if (mask & 0x400000) {
		irqFlags7 |= 0x40;
		if ((IOREGS_ARM7[TM3CNT]&0x4)==0) {
			timer[1][3].enabled = false;
			timer_setup(1,3);
		}
		arm7_irq(0x0000018);
	}


}



unsigned char __stdcall mmu_read_byte(unsigned int address) {
	unsigned char n;

	if ((address>=dtcmBase) && (address<dtcmBase+0x4000))
		__asm {
			mov ebx,address
			mov esi,DTCM
			sub ebx,dtcmBase
			mov al,[esi + ebx]
			mov n,al
		}
	else
		__asm {
			mov ebx,address
			mov edx,arm9_pageTableAddr
			mov edi,arm9_pageLengthAddr
			mov esi,ebx
			mov ecx,arm9_pageMaskAddr
			shr ebx,20			// page
			and esi,0x0FFFFF	// offset
			mov eax,ebx
			shr eax,4

			// ROM reads
			cmp eax,8
			je __rb_8
			cmp eax,9
			jne __rb_maybe_6
			//jne __rb_normal
	__rb_8:
			cmp esi,[edi+eax*4]
			//jge __rb_random
			add esi,[edx+ebx*4]
			mov eax,[esi]
			jmp __rb_done

	__rb_maybe_6:
			cmp eax,0x6
			jne __rb_maybe_A
			xor eax,eax
			jmp __rb_done

			// Save RAM reads
	__rb_maybe_A:
			//cmp eax,0xA
			//jne __rb_normal
			//xor eax,eax
			//jmp __rb_done
			jmp __rb_normal

			// Return "random" data
	__rb_random:
			mov ecx,ebp
			mov eax,address
			ror eax,cl
			jmp __rb_done

	__rb_normal:
			and esi,[ecx+ebx*4]		// AND with pageMask (was eax*4)
			add esi,[edx+ebx*4]		// ADD pageTable[page]
			mov al,[esi]

	__rb_done:
			mov n,al
		}

	return n;
}


unsigned char __stdcall arm7_read_byte(unsigned int address) {
	unsigned char n;

	__asm {
		mov ebx,address
		mov edx,arm7_pageTableAddr
		mov edi,arm7_pageLengthAddr
		mov esi,ebx
		mov ecx,arm7_pageMaskAddr
		shr ebx,20			// page
		and esi,0x0FFFFF	// offset
		mov eax,ebx
		shr eax,4

		// ROM reads
		cmp eax,8
		je __rb_8
		cmp eax,9
		jne __rb_maybe_6
__rb_8:
		cmp esi,[edi+eax*4]
		//jge __rb_random
		add esi,[edx+ebx*4]
		mov eax,[esi]
		jmp __rb_done

__rb_maybe_6:
		cmp eax,0x6
		jne __rb_maybe_A
		xor eax,eax
		jmp __rb_done

		// Save RAM reads
__rb_maybe_A:
		//cmp eax,0xA
		//jne __rb_normal
		//xor eax,eax
		//jmp __rb_done
		jmp __rb_normal

		// Return "random" data
__rb_random:
		mov ecx,ebp
		mov eax,address
		ror eax,cl
		jmp __rb_done

__rb_normal:
		and esi,[ecx+ebx*4]		// AND with pageMask
		add esi,[edx+ebx*4]		// ADD pageTable[page]
		mov al,[esi]

__rb_done:
		//and eax,0xFF
		mov n,al
	}

	return n;
}



/*unsigned char __stdcall mmu_read_byte_old(unsigned int address)
{
	unsigned int page = address>>20,offset; //=address&0xFFFFFF;

		if ((address>=0x237f016)&&(address<0x237f01e))
		{
			wsprintf(buffer,"Rb [%08X]\n",address);
			WriteConsole(hout,buffer,lstrlen(buffer),(unsigned long*)&dummy,NULL);
		}

	if (arm9_pageTable[page])
		//if (pageLength[page]>offset)
			return *(unsigned char*)(arm9_pageTable[page] + ((address&0x0FFFFF)&arm9_pageMask[page]));

	return rand()&0xFF;
}*/




unsigned short int __stdcall mmu_read_word(unsigned int address) {
	unsigned int page = address>>24,pagen=address>>20,offset=address&0x0FFFFF;
	unsigned int tlong;
	unsigned short int *pagew;
	unsigned short int td;


	if (pagen == ipc_page) {
		if (offset == ipc_x)
			return (touch->x*14);
		if (offset== ipc_y)
			return (touch->y*8917)>>9;
		if (offset == ipc_z)
			return (touch->z&0x40)^0x40;
	}

	if (page == 4)
		switch (offset)	{
			case P1:
				return keys;

			case TM0D:
				if (timer[0][0].enabled) {
					tlong = *arm9_cycle;
					if (tlong < timer[0][0].tstart)
						tlong -= (timer[0][0].tstart-0x40000000);
					else
						tlong -= timer[0][0].tstart;
					timer[0][0].tcurr = tlong >> prescalar[IOREGS[TM0CNT]&3];
				}
				return timer[0][0].tcurr;

			case TM1D:
				if (timer[0][1].countup)
					return timer[0][1].te;
				if (timer[0][1].enabled) {
					tlong = *arm9_cycle;
					if (tlong < timer[0][1].tstart)
						tlong -= (timer[0][1].tstart-0x40000000);
					else
						tlong -= timer[0][1].tstart;
					timer[0][1].tcurr = tlong >> prescalar[IOREGS[TM1CNT]&3];
				}
				return timer[0][1].tcurr;

			case TM2D:
				if (timer[0][2].countup)
					return timer[0][2].te;
				if (timer[0][2].enabled) {
					tlong = *arm9_cycle;
					if (tlong < timer[0][2].tstart)
						tlong -= (timer[0][2].tstart-0x40000000);
					else
						tlong -= timer[0][2].tstart;
					timer[0][2].tcurr = tlong >> prescalar[IOREGS[TM2CNT]&3];
				}
				return timer[0][2].tcurr;

			case TM3D:
				if (timer[0][3].countup)
					return timer[0][3].te;
				if (timer[0][3].enabled)
				{
					tlong = *arm9_cycle;
					if (tlong < timer[0][3].tstart)
						tlong -= (timer[0][3].tstart-0x40000000);
					else
						tlong -= timer[0][3].tstart;
					timer[0][3].tcurr = tlong >> prescalar[IOREGS[TM3CNT]&3];
				}
				return timer[0][3].tcurr;

			case IF_:
				/*wsprintf(buffer,"Reading IF (%04x), IE=%04X\n",irqFlags&0xFFFF,
					*(unsigned short int*)(arm9_pageTable[pagen] + 0x210));
				WriteConsole(hout,buffer,lstrlen(buffer),(unsigned long*)&dummy,NULL);*/
				return irqFlags;

			case 0x600:
				gpu_get(GPU_3D_COMMAND);
				//wsprintf(buffer,"GXSTAT = %04X\n",*(unsigned short int*)(IOREGS + 0x600));
				//WriteConsole(hout,buffer,lstrlen(buffer),(unsigned long*)&dummy,NULL);
				return *(unsigned short int*)(IOREGS + 0x600);

			case 0x10:
			case 0x12:
			case 0x14:
			case 0x16:
			case 0x18:
			case 0x1A:
			case 0x1C:
			case 0x1E:
				page = arm9_regs[15];
				pagew = (unsigned short int*)arm9_pageTable[page>>20];
				offset = (page&0x0FFFFF)>>1;
				return pagew[offset];

			default:
				pagew = (unsigned short int*)arm9_pageTable[0x40];
				return pagew[offset>>1];
		}


	if (page == 8)
		if (offset >= arm9_pageLength[0x80])
		{
			page = arm9_regs[15]-4;
			pagew = (unsigned short int*)arm9_pageTable[page>>20];
			offset = (page&0x0FFFFF)>>1;
			return pagew[offset];
		}

	if (arm9_pageTable[pagen])
		return *(unsigned short int*)(arm9_pageTable[pagen] + offset);

	return rand()&0xFFFF;
}


unsigned short int __stdcall arm7_read_word(unsigned int address) {
	unsigned int page = address>>24,pagen=address>>20,offset=address&0x0FFFFF;
	unsigned int tlong;
	unsigned short int *pagew;
	unsigned short int td;


	/*if (pagen == ipc_page)
	{
		if (offset == ipc_x)
			return (touch->x*14);
		if (offset== ipc_y)
			return (touch->y*8917)>>9;
		if (offset == ipc_z)
			return (touch->z&0x40)^0x40;
	}*/

	if (page == 4)
		switch (offset)
		{
			case P1:
				return keys;

		/*	case TM0D:
				if (timer[0].enabled)
				{
					tlong = *arm9_cycle;
					if (tlong < timer[0].tstart)
						tlong -= (timer[0].tstart-0x40000000);
					else
						tlong -= timer[0].tstart;
					timer[0].tcurr = tlong >> prescalar[IOREGS[TM0CNT]&3];
				}
				return timer[0].tcurr;

			case TM1D:
				if (timer[1].countup)
					return timer[1].te;
				if (timer[1].enabled)
				{
					tlong = *arm9_cycle;
					if (tlong < timer[1].tstart)
						tlong -= (timer[1].tstart-0x40000000);
					else
						tlong -= timer[1].tstart;
					timer[1].tcurr = tlong >> prescalar[IOREGS[TM1CNT]&3];
				}
				return timer[1].tcurr;

			case TM2D:
				if (timer[2].countup)
					return timer[2].te;
				if (timer[2].enabled)
				{
					tlong = *arm9_cycle;
					if (tlong < timer[2].tstart)
						tlong -= (timer[2].tstart-0x40000000);
					else
						tlong -= timer[2].tstart;
					timer[2].tcurr = tlong >> prescalar[IOREGS[TM2CNT]&3];
				}
				return timer[2].tcurr;

			case TM3D:
				if (timer[3].countup)
					return timer[3].te;
				if (timer[3].enabled)
				{
					tlong = *arm9_cycle;
					if (tlong < timer[3].tstart)
						tlong -= (timer[3].tstart-0x40000000);
					else
						tlong -= timer[3].tstart;
					timer[3].tcurr = tlong >> prescalar[IOREGS[TM3CNT]&3];
				}
				return timer[3].tcurr;*/

			case IF_:
				/*wsprintf(buffer,"Reading IF (%04x), IE=%04X\n",irqFlags&0xFFFF,
					*(unsigned short int*)(arm9_pageTable[pagen] + 0x210));
				WriteConsole(hout,buffer,lstrlen(buffer),(unsigned long*)&dummy,NULL);*/
				return irqFlags7;

			case 0x600:
				gpu_get(GPU_3D_COMMAND);
				//wsprintf(buffer,"GXSTAT = %04X\n",*(unsigned short int*)(IOREGS + 0x600));
				//WriteConsole(hout,buffer,lstrlen(buffer),(unsigned long*)&dummy,NULL);
				return *(unsigned short int*)(IOREGS + 0x600);

			case 0x10:
			case 0x12:
			case 0x14:
			case 0x16:
			case 0x18:
			case 0x1A:
			case 0x1C:
			case 0x1E:
				page = arm7_regs[15];
				pagew = (unsigned short int*)arm7_pageTable[page>>20];
				offset = (page&0x0FFFFF)>>1;
				return pagew[offset];

			default:
				pagew = (unsigned short int*)arm7_pageTable[0x40];
				return pagew[offset>>1];
		}


	if (page == 8)
		if (offset >= arm7_pageLength[0x80]) {
			page = arm7_regs[15]-4;
			pagew = (unsigned short int*)arm7_pageTable[page>>20];
			offset = (page&0x0FFFFF)>>1;
			return pagew[offset];
		}

	if (arm7_pageTable[pagen])
		return *(unsigned short int*)(arm7_pageTable[pagen] + offset);

	return rand()&0xFFFF;
}



int __stdcall mmu_read_dword(unsigned int address) {
	int n;


	if (address == 0x4000600)
		gpu_get(GPU_3D_COMMAND);


	/*if ((address&0xF00000) == 0x0800000)
	{
		wsprintf(buffer,"Reading from %08X\n",address);
		WriteConsole(hout,buffer,lstrlen(buffer),(unsigned long*)&dummy,NULL);
	}*/

	__asm
	{
		mov ebx,address
		mov edx,arm9_pageTableAddr
		mov edi,arm9_pageLengthAddr
		mov esi,ebx
		mov ecx,arm9_pageMaskAddr
		shr ebx,20			// page
		and esi,0x0FFFFF	// offset
		mov eax,ebx
		shr eax,4

		// ROM reads
		cmp eax,8
		jne __rd_maybe_4
		
		cmp esi,[edi+eax*4]
		jge __rd_random
		test esi,3
		jnz __rd_unaligned
		add esi,[edx+ebx*4]
		mov eax,[esi]
		jmp __rd_done

		// I/O reg reads
__rd_maybe_4:
		cmp eax,4
		jne __rd_normal //__rd_maybe_6
		cmp esi,P1
		jne __rd_normal
		mov ebx,[IOREGS + P1CNT]
		movzx eax,keys
		shl ebx,16
		or eax,ebx
		jmp __rd_done

		// VRAM reads
__rd_maybe_6:
/*		cmp ecx,6
		jne __rd_maybe_0
		
		and esi,0x7FFFF
		//cmp esi,0x17FFF
		//jng __rd_offset_ok
		//sub esi,0x8000		// Mirror upper 32kB
__rd_offset_ok:
		test esi,3
		jnz __rd_unaligned
		add esi,[edx+ebx*4]
		mov eax,[esi]
		jmp __rd_done*/

		// DTCM reads
/*__rd_maybe_0:
		cmp ebx,0
		jne __rd_normal
		test esi,0x800000
		jz __rd_random
		sub esi,0x800000
		mov ebx,[DTCM]
		mov eax,[ebx+esi]
		jmp __rd_done*/

		// Unaligned read, rotate 16 bits
__rd_unaligned:
		mov ecx,esi
		and esi,0xFFFFFFFC		// Clear lower 2 bits of offset
		and ecx,3
		add esi,[edx+ebx*4]		// ADD pageTable[page]
		shl ecx,3
		mov eax,[esi]
		ror eax,cl				// Rotate value by (Offset & 3)*8
		jmp __rd_done

		// Return "random" data
__rd_random:
		mov ecx,ebp
		mov eax,address
		ror eax,cl
		jmp __rd_done

__rd_normal:
		test esi,3
		jnz __rd_unaligned
		and esi,[ecx+ebx*4]		// AND with pageMask
		add esi,[edx+ebx*4]		// ADD pageTable[page]
		mov eax,[esi]

__rd_done:
		mov n,eax
	}


	return n;
}


int __stdcall arm7_read_dword(unsigned int address) {
	int n;


	if (address == 0x4000600)
		gpu_get(GPU_3D_COMMAND);

	__asm
	{
		mov ebx,address
		mov edx,arm7_pageTableAddr
		mov edi,arm7_pageLengthAddr
		mov esi,ebx
		mov ecx,arm7_pageMaskAddr
		shr ebx,20			// page
		and esi,0x0FFFFF	// offset
		mov eax,ebx
		shr eax,4

		// ROM reads
		cmp eax,8
		jne __rd_maybe_4
		
		cmp esi,[edi+eax*4]
		jge __rd_random
		test esi,3
		jnz __rd_unaligned
		add esi,[edx+ebx*4]
		mov eax,[esi]
		jmp __rd_done

		// I/O reg reads
__rd_maybe_4:
		cmp eax,4
		jne __rd_normal //__rd_maybe_6
		cmp esi,P1
		jne __rd_normal
		mov ebx,[IOREGS + P1CNT]
		movzx eax,keys
		shl ebx,16
		or eax,ebx
		jmp __rd_done

		// VRAM reads
__rd_maybe_6:

		// Unaligned read, rotate 16 bits
__rd_unaligned:
		mov ecx,esi
		and esi,0xFFFFFFFC		// Clear lower 2 bits of offset
		and ecx,3
		add esi,[edx+ebx*4]		// ADD pageTable[page]
		shl ecx,3
		mov eax,[esi]
		ror eax,cl				// Rotate value by (Offset & 3)*8
		jmp __rd_done

		// Return "random" data
__rd_random:
		mov ecx,ebp
		mov eax,address
		ror eax,cl
		jmp __rd_done

__rd_normal:
		test esi,3
		jnz __rd_unaligned
		and esi,[ecx+ebx*4]		// AND with pageMask
		add esi,[edx+ebx*4]		// ADD pageTable[page]
		mov eax,[esi]

__rd_done:
		mov n,eax
	}


	return n;
}




void __stdcall mmu_write_byte(unsigned int address,int byte) {
	unsigned int page = (address>>24)&0x0F,pagen=address>>20,offset=address&0x0FFFFF;

	/*if (pagen == 0x08)
	{
		wsprintf(buffer,"Writing %02X to %08X from %08X\n",byte&0xFF,address,cpu_regs[15]);
		WriteConsole(hout,buffer,lstrlen(buffer),(unsigned long*)&dummy,NULL);
	}*/

	if ((address>=dtcmBase) && (address<dtcmBase+0x4000)) {
		*(char*)(DTCM + (address-dtcmBase)) = (char)byte;
		return;
	}

	if (arm9_pageTable[pagen]) {
		if ((pagen>=0x30)&&(pagen<0x40)) {
			*(char*)(arm9_pageTable[pagen] + (offset & arm9_pageMask[pagen])) = (char)byte;
		} else {
			if (arm9_pageLength[page]>offset)
				*(char*)(arm9_pageTable[pagen] + offset) = (char)byte;
		}
	}

	if (pagen==0x40)
	{
		//if (offset == IF_)
		//	irqFlags &= ((char)byte ^ 0xFF);

		/*if (offset<=0x308)
		{
			wsprintf(buffer,"[%04X] = %02X\n",offset,byte);
			WriteConsole(hout,buffer,lstrlen(buffer),(unsigned long*)&dummy,NULL);
		}*/

		switch (offset)
		{
			case IF_:
				irqFlags &= ((byte&0xff) ^ 0xFF)|0xFF00;
				*(unsigned short int*)(arm9_pageTable[pagen] + IF_) = irqFlags;
				break;

			case 0x180:
				*(char*)(arm7_pageTable[pagen] + offset) = (char)byte;
				break;

			case 0x240:
			case 0x241:
			case 0x242:
			case 0x243:
			case 0x244:
			case 0x245:
			case 0x246:
				remap_vram(offset-0x240,byte&0xFF);
				//wsprintf(buffer,"VRAMCNT_%c: %08X\n",offset-1023,byte);
				//WriteConsole(hout,buffer,lstrlen(buffer),(unsigned long*)&dummy,NULL);
				break;
			
			case 0x247:
			if (offset == 0x247) {
				IOREGS_ARM7[0x241] = byte&0xFF;
				remap_wram(0);
				wsprintf(buffer,"WRAMCNT = %X (8bit)\n",byte&0xFF);
				WriteConsole(hout,buffer,lstrlen(buffer),(unsigned long*)&dummy,NULL);
			}
			break;

			case 0x248:
			case 0x249:
				remap_vram(offset-0x241,byte&0xFF);
				break;
		}
	}

}


void __stdcall arm7_write_byte(unsigned int address,int byte) {
	unsigned int page = (address>>24)&0x0F,pagen=address>>20,offset=address&0x0FFFFF;

	if (arm7_pageTable[pagen]) {
		if ((pagen>=0x30)&&(pagen<0x40)) {
			*(char*)(arm7_pageTable[pagen] + (offset & arm7_pageMask[pagen])) = (char)byte;
		} else {
			if (arm7_pageLength[page]>offset)
				*(char*)(arm7_pageTable[pagen] + offset) = (char)byte;
		}
	}

	if (pagen==0x40) {
		/*if (offset<=0x308)
		{
			wsprintf(buffer,"[%04X] = %02X\n",offset,byte);
			WriteConsole(hout,buffer,lstrlen(buffer),(unsigned long*)&dummy,NULL);
		}*/


	//wsprintf(buffer,"(7) [%08X] = %02X, at %X\n",address,byte&0xFF,arm7_regs[15]);
	//WriteConsole(hout,buffer,lstrlen(buffer),(unsigned long*)&dummy,NULL);

		switch (offset) {
			case IF_:
				irqFlags7 &= ((byte&0xff) ^ 0xFF)|0xFF00;
				*(unsigned short int*)(arm7_pageTable[pagen] + IF_) = irqFlags7;
				break;

			case 0x180:
				*(char*)(arm9_pageTable[pagen] + offset) = (char)byte;
				break;

			/*case 0x240:
			case 0x241:
			case 0x242:
			case 0x243:
			case 0x244:
			case 0x245:
			case 0x246:
				remap_vram(offset-0x240,byte&0xFF);
				break;
			case 0x248:
			case 0x249:
				remap_vram(offset-0x241,byte&0xFF);
				break;*/
		}
	}
}





void WRHDECL mmu_write_word(unsigned int address,int word)
{
	unsigned int page = address>>24,pagen=address>>20,offset=address&0x0FFFFF;
	unsigned short int *pagew;


	if (arm9_pageTable[pagen])
		if (arm9_pageLength[page]>offset)
			*(unsigned short int*)(arm9_pageTable[pagen] + offset) = (unsigned short int)word;

	/*if (pagen == 0x08)
	{
		wsprintf(buffer,"[%08X] = %04X\n",(pagen<<20)+offset,mmu_read_word(address)&0xFFFF);
		WriteConsole(hout,buffer,lstrlen(buffer),(unsigned long*)&dummy,NULL);
	}*/

	if (page == 4) {

		switch (offset)
		{
			/*case 0x04:
				wsprintf(buffer,"DISPSTAT %04X\n",word);
				WriteConsole(hout,buffer,lstrlen(buffer),(unsigned long*)&dummy,NULL);
				break;

			case 0x08:
			case 0x0A:
			case 0x0C:
			case 0x0E:
				wsprintf(buffer,"BG%dCNT %04X\n",(offset-0x08)>>1,word);
				WriteConsole(hout,buffer,lstrlen(buffer),(unsigned long*)&dummy,NULL);
				break;
			case 0x12:
			case 0x16:
			case 0x1A:
			case 0x1E:
				wsprintf(buffer,"BG%dY %04X\n",(offset-0x12)>>2,word);
				WriteConsole(hout,buffer,lstrlen(buffer),(unsigned long*)&dummy,NULL);
				break;*/

			case REG_DIVCNT:
				cp_div(word&3);
				break;

			case REG_SQRTCNT:
				cp_sqrt(word&3);
				break;
				
			case 0xC6:
				dma_setup(0,1,false);
				break;
			case 0xD2:
				dma_setup(0,2,false);
				break;
			case 0xDE:
				dma_setup(0,3,false);
				break;
			case IF_:
				//wsprintf(buffer,"Writing %04x to IF",word&0xFFFF);
				//zombie(TELL,ZOMBIE_DEBUG_MSG,buffer,0);
				irqFlags &= (word^0xFFFF);
				break;
			case TM0D:
				//timer[0].tcurr = word&0xFFFF;
				break;
			case TM1D:
				//timer[1].tcurr = word&0xFFFF;
				break;
			case TM2D:
				//timer[2].tcurr = word&0xFFFF;
				break;
			case TM3D:
				//timer[3].tcurr = word&0xFFFF;
				break;
			case TM0CNT:
				timer_setup(0,0);
				break;
			case TM1CNT:
				timer_setup(0,1);
				break;
			case TM2CNT:
				timer_setup(0,2);
				break;
			case TM3CNT:
				timer_setup(0,3);
				break;

			default:
				break;
		}
	}
}



void WRWDECL mmu_write_dword(unsigned int address,int dwrd)
{
	unsigned int page = address>>24,pagen=address>>20,offset=address&0x0FFFFC;
	//int *paged = (int*)(pageTable[pagen]+offset);

	if (arm9_pageTable[pagen])
		if (arm9_pageLength[page]>offset)
			*(int*)(arm9_pageTable[pagen] + offset) = dwrd;

	/*if (page == 6)
	{
		wsprintf(buffer,"[%08X] = %08X at %08X\n",address,dwrd,cpu_regs[15]);
		WriteConsole(hout,buffer,lstrlen(buffer),(unsigned long*)&dummy,NULL);
	}*/

	if (pagen != 0x40) return;

	//if (pagen == 0x40)
	//{
		if ((offset>=0x400)&&(offset<=0x600))
		{
			//wsprintf(buffer,"[%04X] = %04X\n",offset&0xFFFF,dwrd);
			//WriteConsole(hout,buffer,lstrlen(buffer),(unsigned long*)&dummy,NULL);
			gpu_set(GPU_3D_COMMAND,offset,dwrd);
		}
		
		switch (offset) 
		{
			/*case 0x600:
				wsprintf(buffer,"3D: %08X (PC=%08X) %04X (%08X)\n",dwrd,cpu_regs[15],pageLength[page],*paged);
				WriteConsole(hout,buffer,lstrlen(buffer),(unsigned long*)&dummy,NULL);
				break;
			/*case 0x1000:
				wsprintf(buffer,"DISPCNT_S: %08X\n",dwrd);
				WriteConsole(hout,buffer,lstrlen(buffer),(unsigned long*)&dummy,NULL);
				break;*/
			/*case 0x00:
				wsprintf(buffer,"DISPCNT %08X\n",dwrd);
				WriteConsole(hout,buffer,lstrlen(buffer),(unsigned long*)&dummy,NULL);
				break;

			case 0x240:
			case 0x241:
			case 0x242:
			case 0x243:
				wsprintf(buffer,"VRAMCNT_%c: %08X\n",offset-175,dwrd);
				WriteConsole(hout,buffer,lstrlen(buffer),(unsigned long*)&dummy,NULL);
				break;*/

			case REG_NUMER_L:
			case REG_DENOM_L:
				cp_div(-1);
				break;
			case REG_SQRTPARM:
				cp_sqrt(-1);
				break;

			case TM0D:
				timer_setup(0,0);
				break;
			case TM1D:
				timer_setup(0,1);
				break;
			case TM2D:
				timer_setup(0,2);
				break;
			case TM3D:
				timer_setup(0,3);
				break;

			//case IE:
				//wsprintf(buffer,"Writing %08x to IE",dwrd);
				//zombie(TELL,ZOMBIE_DEBUG_MSG,buffer,0);
			//	irqFlags ^= (dwrd>>16);
			//	break;
			case IF_:
				//wsprintf(buffer,"Writing %08x to IF",dwrd);
				//zombie(TELL,ZOMBIE_DEBUG_MSG,buffer,0);
				irqFlags &= (dwrd^0xFFFF);
				break;

			case 0xC4:
				dma_setup(0,1,false);
				break;
			case 0xC6:
				dma_setup(0,1,false);
				break;
			case 0xD0:
				dma_setup(0,2,false);
				break;
			case 0xD2:
				dma_setup(0,2,false);
				break;
			/*case 0xD4:
				wsprintf(buffer,"DM3SAD = %08x\n",dwrd);
				WriteConsole(hout,buffer,lstrlen(buffer),(unsigned long*)&dummy,NULL);
				break;*/
			case 0xD8:
				//wsprintf(buffer,"DM3DAD = %08X (PC = %08X)\n",dwrd,cpu_regs[15]);
				//WriteConsole(hout,buffer,lstrlen(buffer),(unsigned long*)&dummy,NULL);
				//wsprintf(buffer,"%08X: %08X %08X %08X\n",cpu_regs[13],
				//	mmu_read_dword(cpu_regs[13]),mmu_read_dword(cpu_regs[13]+4),mmu_read_dword(cpu_regs[13]+8));
				//WriteConsole(hout,buffer,lstrlen(buffer),(unsigned long*)&dummy,NULL);
				break;
			case 0xDC:
				dma_setup(0,3,false);
				break;
			case 0xDE:
				dma_setup(0,3,false);
				break;
			default:
				break;
		}
//	}
}



int __stdcall mmu_get(int what)
{
	if (what&0x10000) {
		what &= 0xFFFF;
		//	wsprintf(buffer,"mmu_get_%d(%x)\n",1,what);
		//	WriteConsole(hout,buffer,lstrlen(buffer),(unsigned long*)&dummy,NULL);

		switch (what) {
			case MMU_IRQ_OK:
				return (irqFlags&0xFFFF)&irqAck;
			case MMU_READ_BYTE_CALLBACK:
				return (int)arm7_read_byte;
			case MMU_READ_BYTE_CALLBACKS:
				return (int)&arm7_rd8[0];
			case MMU_READ_WORD_CALLBACK:
				return (int)arm7_read_word;
			case MMU_READ_WORD_CALLBACKS:
				return (int)&arm7_rd16[0];
			case MMU_READ_DWORD_CALLBACKS:
				return (int)&arm7_rd32[0];
			case MMU_READ_DWORD_CALLBACK:
				return (int)arm7_read_dword;
			case MMU_WRITE_BYTE_CALLBACK:
				return (int)arm7_write_byte;
			case MMU_WRITE_BYTE_CALLBACKS:
				return (int)&arm7_wr8[0];
			case MMU_WRITE_WORD_CALLBACK:
				return (int)mmu_write_word;
			case MMU_WRITE_WORD_CALLBACKS:
				return (int)&arm7_wr16[0];
			case MMU_WRITE_DWORD_CALLBACKS:
				return (int)&arm7_wr32[0];
			case MMU_WRITE_DWORD_CALLBACK:
				return (int)mmu_write_dword;
			case MMU_TRANSLATE_ADDRESS_CALLBACK:
				return (int)arm7_translate_address;
			case MMU_IRQ_CALLBACK:
				return (int)mmu_irq;
			case MMU_GET_CALLBACK:
				return (int)mmu_get;
			case MMU_PAGE_TABLE:
				return (int)&arm7_pageTable[0];
			case MMU_PAGE_MASKS:
				return (int)&arm7_pageMask[0];
			case MMU_VRAM_BANKS:
				return (int)&vramBank[0];
			default:
				break;
		}
	} else {
		//	wsprintf(buffer,"mmu_get_%d(%x)\n",0,what);
		//	WriteConsole(hout,buffer,lstrlen(buffer),(unsigned long*)&dummy,NULL);

		switch (what) {
			case MMU_DISP_FIFO:
				int *dmareg;
				dmareg = (int*)(arm9_pageTable[0x40]+0xB0);
				for (int i=0; i<4; i++)	{
					if (dmareg[(i*3)+2] == 0xA6400004) {
						if ((dmareg[(i*3)+1] == 0x04000068) &&
							((dmareg[(i*3)]&0xFF000000) == 0x02000000)) {
							return (int)mmu_translate_address(dmareg[(i*3)]);
						}
					}
					
				}
				return 0;

			case MMU_DTCM:
				return (int)DTCM;
			case MMU_ITCM:
				return (int)ITCM;
			case MMU_IRQ_OK:
				return (irqFlags&0xFFFF)&irqAck;
			case MMU_READ_BYTE_CALLBACK:
				return (int)mmu_read_byte;
			case MMU_READ_WORD_CALLBACK:
				return (int)mmu_read_word;
			case MMU_READ_BYTE_CALLBACKS:
				return (int)&arm9_rd8[0];
			case MMU_READ_WORD_CALLBACKS:
				return (int)&arm9_rd16[0];
			case MMU_READ_DWORD_CALLBACKS:
				return (int)&arm9_rd32[0];
			case MMU_READ_DWORD_CALLBACK:
				return (int)mmu_read_dword;
			case MMU_WRITE_BYTE_CALLBACK:
				return (int)mmu_write_byte;
			case MMU_WRITE_WORD_CALLBACK:
				return (int)mmu_write_word;
			case MMU_WRITE_BYTE_CALLBACKS:
				return (int)&arm9_wr8[0];
			case MMU_WRITE_WORD_CALLBACKS:
				return (int)&arm9_wr16[0];
			case MMU_WRITE_DWORD_CALLBACKS:
				return (int)&arm9_wr32[0];
			case MMU_WRITE_DWORD_CALLBACK:
				return (int)mmu_write_dword;
			case MMU_TRANSLATE_ADDRESS_CALLBACK:
				return (int)mmu_translate_address;
			case MMU_IRQ_CALLBACK:
				return (int)mmu_irq;
			case MMU_GET_CALLBACK:
				return (int)mmu_get;
			case MMU_PAGE_TABLE:
				return (int)&arm9_pageTable[0];
			case MMU_PAGE_MASKS:
				return (int)&arm9_pageMask[0];
			case MMU_VRAM_BANKS:
				return (int)&vramBank[0];
			case PLUGIN_NAME:
				return (int)&mmuName;
			case PLUGIN_VERSION:
				return 25;
			case 0x700:
				return (int)mmu_translate_address((unsigned int)0x8000288);
			default:
				break;
		}
	}
	return 0;
}


void __stdcall mmu_set(int what,void* param1,void* param2) {
	static char buffer[64];
	int i,j,n,proc;
	unsigned short int s;
	unsigned char *iobase;

	proc = 0;

	switch (what) {
		case CPU_CYCLE_PTR:
			if ((int)param2 == 0)
				arm9_cycle = (unsigned int *)param1;
			else
				arm7_cycle = (unsigned int *)param1;
			break;

		case MMU_TIMER_OVERFLOW:
			n = ((int)param1)>>16;
			proc = (n&4)>>2;
			n &= 3;

			//wsprintf(buffer,"Overflow on timer %d (cnt=%08X)\n",n,*(int*)(&IOREGS_ARM7[TM0CNT+n*4]));
			//WriteConsole(hout,buffer,lstrlen(buffer),(unsigned long*)&dummy,NULL);

			if (proc==0) {
				if (IOREGS[TM0CNT+n*4]&0x40)
					mmu_irq(0x08<<n);
				else
					timer_setup(proc,n);
			} else {
				if (IOREGS_ARM7[TM0CNT+n*4]&0x40)
					mmu_irq((0x08<<n) << 16);
				else
					timer_setup(proc,n);
			}
			if (n<3)
				if (proc == 0) {
					if (IOREGS[TM0CNT+((n+1)*4)]&0x04)
					{
						//wsprintf(buffer,"Doing count-up on timer %d (%d)\n",n+1,timer[n+1].tcurr+1);
						//WriteConsole(hout,buffer,lstrlen(buffer),(unsigned long*)&dummy,NULL);

						timer[0][n+1].te++;
						if (timer[0][n+1].te==0) {
							//wsprintf(buffer,"Timer %d overflowed, reloading with %04x\n",n+1,timer[n+1].td);
							//WriteConsole(hout,buffer,lstrlen(buffer),(unsigned long*)&dummy,NULL);

							timer[0][n+1].te = timer[0][n+1].td;
							if (IOREGS[TM0CNT+((n+1)*4)]&0x40)
								mmu_irq(0x08<<(n+1));
						}
					}
				} else {
					if (IOREGS_ARM7[TM0CNT+((n+1)*4)]&0x04) {
						timer[1][n+1].te++;
						if (timer[1][n+1].te==0) {
							//wsprintf(buffer,"Timer %d overflowed, reloading with %04x\n",n+1,timer[1][n+1].td);
							//WriteConsole(hout,buffer,lstrlen(buffer),(unsigned long*)&dummy,NULL);

							timer[1][n+1].te = timer[1][n+1].td;
							if (IOREGS_ARM7[TM0CNT+((n+1)*4)]&0x40)
								mmu_irq((0x08<<(n+1)) << 16);
						}
					}
				}
			break;

		case MMU_TIMER_COUNTUP:
			n = ((int)param1)>>16;
			proc = (n&4)>>2;
			n &= 3;
			iobase = (proc==0)?IOREGS:IOREGS_ARM7;

			//wsprintf(buffer,"Overflow on timer %d,%d (unchanged=%d)\n",proc,n,timer[proc][n]);
			//WriteConsole(hout,buffer,lstrlen(buffer),(unsigned long*)&dummy,NULL);

			if (timer[proc][n].unchanged) {
				timer[proc][n].tstart = (proc==0)?(*arm9_cycle):(*arm7_cycle);
				timer[proc][n].tstart <<= proc;
				zombie(TELL,ZOMBIE_EVENT,(void*)(EV_TIMER_COUNTUP|((n+proc*4)<<16)),
					   (void*)(timer[proc][n].tstart + timer[proc][n].interval));
			} else {
				timer_setup(proc,n);
			}
			
			//if (iobase[TM0CNT+((n+1)*4)]&0x04) {
			timer[proc][n+1].te++;
			if (timer[proc][n+1].te == 0) {
				timer[proc][n+1].te = timer[proc][n+1].td;
				if (iobase[TM0CNT+((n+1)*4)]&0x40)
					mmu_irq(0x08<<(n+1+(proc*16)));
			}
			//}
			break;


		case MMU_IRQ:
			if ((int)param2==0) {
				if (IOREGS[0x208]&1) {
					irqFlags |= (int)param1;
					*(unsigned short*)(arm9_pageTable[0x40] + 0x214) = irqFlags;
				}
			} else {
				if (IOREGS_ARM7[0x208]&1) {
					irqFlags7 |= (int)param1;
					*(unsigned short*)(arm7_pageTable[0x40] + 0x214) = irqFlags7;
				}
			}

			irqMask = irqFlags;
			break;

		case MMU_IRQ_OVER:
			s = *(unsigned short int*)(arm9_pageTable[dtcmPage] + 0x3FF8);

			//wsprintf(buffer,"ACK: %04X, MSK: %04X, s: %04X",irqAck,irqMask,s);
			//zombie(TELL,ZOMBIE_DEBUG_MSG,buffer,0);

			if ((irqMask & s) != irqMask)
				irqAck &= ((irqMask^0xFFFF)|s);
			//*(unsigned short int*)(pageTable[dtcmPage] + 0x3FF8) = 0;
			break;

		case MMU_TCM_REMAP:
			if (param1==0)
			{
				//pageTable[dtcmPage<<4] = 0;
				dtcmBase = (unsigned int)param2&0xFFFFF000;
				if (((int)param2>>20) != dtcmPage) {
					if (dtcmLast.nPage != -1) {
						arm9_wr8[dtcmPage]  = dtcmLast.wrb;
						arm9_rd8[dtcmPage]  = dtcmLast.rdb;
						arm9_wr16[dtcmPage] = dtcmLast.wrh;
						arm9_rd16[dtcmPage] = dtcmLast.rdh;
						arm9_wr32[dtcmPage] = dtcmLast.wrw;
						arm9_rd32[dtcmPage] = dtcmLast.rdw;
						arm9_pageTable[dtcmPage] = dtcmLast.physaddr;
						arm9_pageLength[dtcmPage] = dtcmLast.length;
						for (i=0; i<16; i++)
							arm9_pageMask[(dtcmPage<<4)+i] = dtcmLast.mask;
					}
					dtcmPage = (int)param2>>20;

					dtcmLast.nPage = dtcmPage;
					dtcmLast.wrb = arm9_wr8[dtcmPage];
					dtcmLast.rdb = arm9_rd8[dtcmPage];
					dtcmLast.wrh = arm9_wr16[dtcmPage];
					dtcmLast.rdh = arm9_rd16[dtcmPage];
					dtcmLast.wrw = arm9_wr32[dtcmPage];
					dtcmLast.rdw = arm9_rd32[dtcmPage];
					dtcmLast.physaddr = arm9_pageTable[dtcmPage];
					dtcmLast.length = arm9_pageLength[dtcmPage];
					dtcmLast.mask = arm9_pageMask[dtcmPage<<4];

					arm9_rd8[dtcmPage]  = (mmu_read_byte_callback)arm9_read_byte_dtcm;
					arm9_rd16[dtcmPage] = (mmu_read_word_callback)arm9_read_word_dtcm;
					arm9_rd32[dtcmPage] = (mmu_read_dword_callback)arm9_read_dword_dtcm;
					arm9_wr8[dtcmPage]  = (mmu_write_byte_callback)arm9_write_byte_dtcm;
					arm9_wr16[dtcmPage] = (mmu_write_word_callback)arm9_write_word_dtcm;
					arm9_wr32[dtcmPage] = (mmu_write_dword_callback)arm9_write_dword_dtcm;
					//arm9_pageTable[dtcmPage] = DTCM;
					//arm9_pageMask[dtcmPage] = 0x3FFF;

					wsprintf(buffer,"Remapping DTCM to 0x%08X",(int)param2&0xFFFFF000);
					zombie(TELL,ZOMBIE_DEBUG_MSG,buffer,0);
					//pageTable[(int)param2>>20] = DTCM;
					//pageLength[(int)param2>>24] = 0x4000;
					//pageMask[(int)param2>>24] = 0x3FFF;
				}
			}
			break;

		case MMU_LZ77_UNPACK_ARM7:
			proc = 1;
		case MMU_LZ77_UNPACK_ARM9:
			UINT dest,src,src2,*cycptr;
			UINT N,size,disp,dataHdr;
			unsigned char flags;

			mmu_read_byte_callback *rdb;
			mmu_write_byte_callback *wrb;

			dest = (UINT)param1;
			src  = (UINT)param2;
			rdb  = (proc==0)?(&arm9_rd8[0]):(&arm7_rd8[0]);
			wrb  = (proc==0)?(&arm9_wr8[0]):(&arm7_wr8[0]);
			cycptr = (proc==0)?arm9_cycle:arm7_cycle;

			//wsprintf(buffer,"Unpacking from %X to %X on processor %d\n",src,dest,proc);
			//zombie(TELL,ZOMBIE_DEBUG_MSG,buffer,0);
			//WriteConsole(hout,buffer,lstrlen(buffer),(unsigned long*)&dummy,NULL);

			if (proc == 0)
				dataHdr = arm9_rd32[src>>20](src);
			else
				dataHdr = arm7_rd32[src>>20](src);
			src += 4;
			if ((dataHdr&0xF0)!=0x10) return;
			size = (dataHdr>>8);

			//wsprintf(buffer,"Unpacking %d bytes from %X to %X (%X)\n",size,src-4,dest,dataHdr);
			//zombie(TELL,ZOMBIE_DEBUG_MSG,buffer,0);
			//WriteConsole(hout,buffer,lstrlen(buffer),(unsigned long*)&dummy,NULL);

			while (size > 0) {
				flags = rdb[src>>20](src); src++;
				if (flags) {
					for (i=0; i<8; i++) {
						if (flags & 0x80) {
							N = rdb[src>>20](src); src++;
							disp = rdb[src>>20](src); src++;
							disp |= (N&0xF)<<8;
							src2 = dest - disp - 1;
							N = (N>>4)+3;
							for (j=0; j<N; j++) {
								wrb[dest>>20](dest,rdb[src2>>20](src2));
								src2++; dest++;
								if ((--size) == 0) return;
							}
							*cycptr += N*4;
						} else {
							wrb[dest>>20](dest,rdb[src>>20](src));
							src++; dest++;
							if ((--size) == 0) return;
							*cycptr += 4;
						}
						flags <<= 1;
					}
				} else {
					for (i=0; i<8; i++) {
						wrb[dest>>20](dest,rdb[src>>20](src));
						src++; dest++;
						if ((--size) == 0) return;
					}
					*cycptr += 32;
				}
			}

			break;

		case ZOMBIE_KEYSTATE:
			if ((int)param2==1)
			{
				/*for (i=0; i<10; i++)
					if ((int)param1 == keycodes[i])
					{
						keys &= (keymask[i] ^ 0x3FF);
						break;
					}*/
				i = (int)param1;
				if ((i>=0)&&(i<10))
					keys &= (keymask[i] ^ 0x3FF);
				else if ((i>=10) && (i<12))
					btnXY[i-10] = 1;	

			} else
			{
				/*for (i=0; i<10; i++)
					if ((int)param1 == keycodes[i])
					{
						keys |= (keymask[i]);
						break;
					}*/
				i = (int)param1;
				if ((i>=0)&&(i<10))
					keys |= keymask[i];
				else if ((i>=10) && (i<12))
					btnXY[i-10] = 0;	


			}
			if (!IOREGS) return;

			if (IOREGS[IME]&1)
			if (IOREGS[IE+1]&0x10)
				if (IOREGS[P1CNT+1]&0x40)
					if (IOREGS[P1CNT+1]&0x80)
					{
					// Logical multiplication (AND)
						if (((keys^0xFFFF)&0x3FF) == (IOREGS[P1CNT]&0x3FF))
						{
							//IOREGS[0x201] |= 0x20;
							irqFlags |= 0x1000;
							cpu_irq(0x00000018);
							//arm7_irq(0x18);
						}
					}
					//else if ((keys&IOREGS[P1CNT])!=IOREGS[P1CNT])
					// Logical addition (OR)
					else if ((keys^0xFFFF)&(IOREGS[P1CNT]&0x3FF))
					{
						irqFlags |= 0x1000;
						cpu_irq(0x00000018);
						//arm7_irq(0x18);
					}
			break;

		case PLUGIN_SETTING:
			switch ((int)param1)
			{
				case 0:
				case 1:
				case 2:
				case 3:
				case 4:
				case 5:
				case 6:
				case 7:
				case 8:
				case 9:
					keycodes[(int)param1] = (int)param2;
					break;
				
				case 10:
					arm9_rd16[ipc_page] = (mmu_read_word_callback)arm9_read_word_gen;

					ipc_page = 0x20 + ((int)param2*3);
					ipc_x = 0x04 + ((int)param2*0x7f000);	// 0x7f000
					ipc_y = 0x06 + ((int)param2*0x7f000);
					ipc_z = 0x14 + ((int)param2*0x7f000);
					arm9_rd16[ipc_page] = rdh_ipc;
					//arm9_rd16[0x27] = arm9_rd16[0x23];
					break;
				
				case 11:
					if (param2)	{
						arm9_rd16[0x27] = rdh_ipc_mirror;
					} else {
						arm9_rd16[0x27] = (mmu_read_word_callback)arm9_read_word_gen;
					}
					zombie(TELL, MMU_READ_WORD_CALLBACKS, (void*)arm9_rd16, 0);
					break;

				case 20:
					if ((int)param2 == 1) {
						rdh_ipc = (mmu_read_word_callback)arm9_read_word_ipc;
						rdh_ipc_mirror = (mmu_read_word_callback)arm9_read_word_ipc_mirror;

						arm9_rd16[ipc_page] = (mmu_read_word_callback)arm9_read_word_ipc;
						if (arm9_rd16[0x27] != (mmu_read_word_callback)arm9_read_word_gen)
							arm9_rd16[0x27] = (mmu_read_word_callback)arm9_read_word_ipc_mirror;
					} else if ((int)param2 == 2) {
						rdh_ipc = (mmu_read_word_callback)arm9_read_word_ipc_v2;
						rdh_ipc_mirror = (mmu_read_word_callback)arm9_read_word_ipc_mirror_v2;

						arm9_rd16[ipc_page] = (mmu_read_word_callback)arm9_read_word_ipc_v2;
						if (arm9_rd16[0x27] != (mmu_read_word_callback)arm9_read_word_gen)
							arm9_rd16[0x27] = (mmu_read_word_callback)arm9_read_word_ipc_mirror_v2;
					}
					zombie(TELL, MMU_READ_WORD_CALLBACKS, (void*)arm9_rd16, 0);
					break;

				case 60:
					logDMA = (int)param2;
					break;

				case 63:
					blockDMA = 1-(int)param2;
					break;

				case 70:
					cfDevice = (int)param2;
					break;

				default:
					break;
			}
			break;

		default:
			break;
	}
}



// Configuration dialog callback
BOOL CALLBACK DlgProc(HWND hdlg, UINT msg, WPARAM wParam, LPARAM lParam) {
	static char stemp[64];
	int i;
	HWND hCmb,hChk;

   switch(msg) {
		case WM_INITDIALOG:
			SetWindowLong(hdlg, DWL_USER, lParam);

			hChk = GetDlgItem(hdlg,IDC_BIOS);
			SendMessage(hChk, BM_SETCHECK, (useExtBios)?BST_CHECKED:BST_UNCHECKED, 0);

			hChk = GetDlgItem(hdlg,IDC_EDIT1);
			SetWindowText(hChk,szBiosName);
			EnableWindow(hChk,0);

			return TRUE;

		case WM_COMMAND:
			switch(LOWORD(wParam)) {
				case IDOK:

					EndDialog(hdlg, 0);
					return TRUE;

				case IDCANCEL:
					EndDialog(hdlg, 1);
					return FALSE;

				default:
					break;
			}
			break;
   }

	return FALSE;
}

